import bpy
import pytest
import MDAnalysis as mda
import numpy as np
import molecularnodes as mn
from molecularnodes.entities.trajectory import dna
from databpy.object import LinkedObjectError
from .utils import NumpySnapshotExtension
from .constants import data_dir


class TestOXDNAReading:
    @pytest.fixture(scope="module")
    def file_holl_top(self):
        return data_dir / "oxdna/holliday.top"

    @pytest.fixture(scope="module")
    def file_top_new(self):
        return data_dir / "oxdna/top_new.top"

    @pytest.fixture(scope="module")
    def file_top_new_custom(self):
        return data_dir / "oxdna/top_new_custom.top"

    @pytest.fixture(scope="module")
    def file_top_old(self):
        return data_dir / "oxdna/top_old.top"

    @pytest.fixture(scope="module")
    def file_traj_old_new(self):
        return data_dir / "oxdna/traj_old_new.dat"

    @pytest.fixture(scope="module")
    def file_holl_dat(self):
        return data_dir / "oxdna/holliday.dat"

    @pytest.fixture(scope="module")
    def universe(self, file_holl_top, file_holl_dat):
        return mda.Universe(
            file_holl_top,
            file_holl_dat,
            format=dna.OXDNAReader,
            topology_format=dna.OXDNAParser,
        )

    def test_read_as_universe(self, file_holl_top, file_holl_dat):
        u = mda.Universe(
            file_holl_top,
            file_holl_dat,
            format=dna.OXDNAReader,
            topology_format=dna.OXDNAParser,
        )
        assert u.atoms.n_atoms == 98

    def test_univ_as_traj(self, universe):
        traj = dna.OXDNA(universe)
        assert traj.universe
        with pytest.raises(LinkedObjectError):
            traj.object
        assert all([x in ["A", "C", "T", "G"] for x in traj.res_name])

    def test_univ_snapshot(self, universe: mda.Universe, snapshot_custom):
        traj = dna.OXDNA(universe)
        traj.create_object()
        for name in ["position", "res_name", "res_id", "chain_id"]:
            assert snapshot_custom == getattr(traj, name)

    def test_detect_new_top(self, file_top_old, file_top_new, file_top_new_custom):
        assert dna.OXDNAParser._is_new_topology(file_top_new)
        assert dna.OXDNAParser._is_new_topology(file_top_new_custom)
        assert not dna.OXDNAParser._is_new_topology(file_top_old)

    def test_topo_reading(
        self,
        file_top_old,
        file_top_new,
        file_top_new_custom,
    ):
        top_new = dna.OXDNAParser._read_topo_new(file_top_new)
        top_new_custom = dna.OXDNAParser._read_topo_new(file_top_new_custom)
        top_old = dna.OXDNAParser._read_topo_old(file_top_old)

        for top in [top_new, top_old, top_new_custom]:
            assert top.n_atoms == 12
            assert top.n_residues == 12

    @pytest.mark.parametrize("topfile", ["top_new", "top_new_custom", "top_old"])
    def test_comparing_topologies(self, snapshot, topfile, file_traj_old_new):
        u = mda.Universe(
            data_dir / f"oxdna/{topfile}.top",
            file_traj_old_new,
            topology_format=dna.OXDNAParser,
            format=dna.OXDNAReader,
        )
        traj = dna.OXDNA(u)
        traj.create_object()
        assert len(traj) == 12
        assert snapshot == traj.bonds.tolist()
        for att in ["res_id", "chain_id", "res_name"]:
            assert snapshot == traj.named_attribute(att).tolist()

    def test_reading_example(self):
        traj = dna.OXDNA(
            mda.Universe(
                data_dir / "CanDo2oxDNA/top.top",
                data_dir / "CanDo2oxDNA/traj.oxdna",
                topology_format=dna.OXDNAParser,
                format=dna.OXDNAReader,
            )
        )
        traj.create_object()
        assert len(np.unique(traj.named_attribute("res_id"))) == 15166
        assert len(np.unique(traj.named_attribute("chain_id"))) == 178

    def test_session_register(self, file_holl_top, file_holl_dat):
        session = mn.session.get_session()
        u = mda.Universe(
            file_holl_top,
            file_holl_dat,
            topology_format=dna.OXDNAParser,
            format=dna.OXDNAReader,
        )
        traj = dna.OXDNA(u)
        traj.create_object()

        assert isinstance(session.get(traj.uuid), dna.OXDNA)

    def test_reload_lost_connection(self, snapshot, file_holl_top, file_holl_dat):
        session = mn.session.get_session()
        u = mda.Universe(
            file_holl_top,
            file_holl_dat,
            topology_format=dna.OXDNAParser,
            format=dna.OXDNAReader,
        )
        traj = dna.OXDNA(u)
        traj.create_object()
        obj_name = traj.name
        bpy.context.scene.frame_set(1)
        pos1 = traj.position
        bpy.context.scene.frame_set(2)
        pos2 = traj.position
        assert not np.allclose(pos1, pos2)

        traj_old = session.entities.pop(traj.uuid)

        # the position shouldn't change as we have removed the traj from the session
        bpy.context.scene.frame_set(3)
        pos3 = traj.position
        assert np.allclose(pos2, pos3)
        del traj

        bpy.data.objects[obj_name].select_set(True)
        bpy.ops.mn.reload_trajectory()

        # when reloading the object, a brand new traj had to be created, which updates
        # the uuid on the object, so the old traj will not longer be able to find any
        # matching object and instead we'll have to look back up a new traj based on the
        # the object's uuid
        with pytest.raises(LinkedObjectError):
            traj_old.object.name

        traj = bpy.context.scene.MNSession.get(bpy.context.active_object.uuid)
        assert traj is not None

        pos3 = traj.position

        assert not np.allclose(pos2, pos3)



import molecularnodes as mn

import random
from .constants import data_dir
from .utils import NumpySnapshotExtension


def test_ss_label_to_int():
    examples = ["TURN_TY1_P68", "BEND64", "HELX_LH_PP_P9", "STRN44"]
    assert [3, 3, 1, 2] == [
        mn.entities.molecule.pdbx._ss_label_to_int(x) for x in examples
    ]


def test_entity_parsing():
    mn.entities.fetch("6VBU", format="bcif")
    assert True


def test_get_ss_from_mmcif(snapshot_custom: NumpySnapshotExtension):
    mol = mn.entities.load_local(data_dir / "1cd3.cif")

    # mol2, fil2 = mn.io.fetch('1cd3')

    random.seed(6)
    random_idx = random.sample(range(len(mol)), 100)

    # assert (mol.sec_struct == mol2.sec_struct)[random_idx].all()

    assert snapshot_custom == mol.array.sec_struct[random_idx]


def test_secondary_structure_no_helix(snapshot_custom):
    m = mn.entities.fetch("7ZL4", cache_dir=data_dir)
    assert snapshot_custom == m.named_attribute("sec_struct")



import bpy
from bpy.types import Collection
from databpy.collection import create_collection


def mn() -> Collection:
    "Return the 'MolecularNodes' collection, creating it first if required"
    return create_collection("MolecularNodes")


def data() -> Collection:
    "Return the MolecularNodes/data collection and disable it"
    name = ".MN_data"

    try:
        return bpy.data.collections[name]
    except KeyError:
        collection = create_collection(name=name, parent=mn())

        bpy.context.view_layer.layer_collection.children["MolecularNodes"].children[
            collection.name
        ].exclude = True
        return collection


def frames(name: str = "") -> Collection:
    "Return a collection for storing the objects that are the frames of a trajectory"
    return create_collection(f".data_{name}_frames", parent=data())


def cellpack(name: str = "") -> Collection:
    "Return a collection for storing the instances for a CellPack Ensemble"
    full_name = f"cellpack_{name}"
    return create_collection(full_name, parent=data())



import bpy
from bpy.types import PropertyGroup
from bpy.props import IntProperty, BoolProperty, EnumProperty, StringProperty
from .handlers import _update_entities
from .style import STYLE_ITEMS
from .session import get_session

uuid_property = StringProperty(  # type: ignore
    name="UUID",
    description="Unique ID for referencing the required objects in the MNSession",
    default="",
)


class MolecularNodesSceneProperties(PropertyGroup):
    import_del_hydrogen: BoolProperty(  # type: ignore
        name="Remove Hydrogens",
        description="Remove the hydrogens from a structure on import",
        default=False,
    )

    import_local_path: StringProperty(  # type: ignore
        name="File",
        description="File path of the structure to open",
        options={"TEXTEDIT_UPDATE"},
        subtype="FILE_PATH",
        maxlen=0,
    )

    import_code_alphafold: StringProperty(  # type: ignore
        name="UniProt ID",
        description="The UniProt ID to use for downloading from the AlphaFold databse",
        options={"TEXTEDIT_UPDATE"},
    )

    import_format_alphafold: EnumProperty(  # type: ignore
        name="Format",
        description="Format to download as from the PDB",
        items=(
            # ("bcif", ".bcif", "Binary compressed .cif file, fastest for downloading"),
            ("cif", ".cif", "The new standard of .cif / .mmcif"),
            ("pdb", ".pdb", "The classic (and depcrecated) PDB format"),
        ),
    )

    import_format_wwpdb: EnumProperty(  # type: ignore
        name="Format",
        description="Format to download as from the PDB",
        items=(
            ("bcif", ".bcif", "Binary compressed .cif file, fastest for downloading"),
            ("cif", ".cif", "The new standard of .cif / .mmcif"),
            ("pdb", ".pdb", "The classic (and depcrecated) PDB format"),
        ),
    )

    import_code_pdb: StringProperty(  # type: ignore
        name="PDB",
        description="The PDB code to download and import",
        options={"TEXTEDIT_UPDATE"},
        maxlen=4,
    )

    is_updating: BoolProperty(  # type: ignore
        name="Updating",
        description="Currently updating data in the scene, don't trigger more updates",
        default=False,
    )

    import_centre: BoolProperty(  # type: ignore
        name="Centre Structure",
        description="Move the imported Molecule on the World Origin",
        default=False,
    )

    import_centre_type: EnumProperty(  # type: ignore
        name="Method",
        default="mass",
        items=(
            (
                "mass",
                "Mass",
                "Adjust the structure's centre of mass to be at the world origin",
                1,
            ),
            (
                "centroid",
                "Centroid",
                "Adjust the structure's centroid (centre of geometry) to be at the world origin",
                2,
            ),
        ),
    )

    import_node_setup: BoolProperty(  # type: ignore
        name="Setup Nodes",
        default=True,
        description="Create and set up a Geometry Nodes tree on import",
    )

    import_build_assembly: BoolProperty(  # type: ignore
        name="Build Assembly",
        description="Add a node to build the biological assembly on import",
        default=False,
    )

    import_del_solvent: BoolProperty(  # type: ignore
        name="Remove Solvent",
        description="Delete the solvent from the structure on import",
        default=True,
    )

    import_oxdna_topology: StringProperty(  # type: ignore
        name="Toplogy",
        description="File path for the topology to import (.top)",
        subtype="FILE_PATH",
    )
    import_oxdna_trajectory: StringProperty(  # type: ignore
        name="Trajectory",
        description="File path for the trajectory to import (.oxdna / .dat)",
        subtype="FILE_PATH",
    )
    import_oxdna_name: StringProperty(  # type: ignore
        name="Name", description="Name of the created object.", default="NewOrigami"
    )
    import_style: EnumProperty(  # type: ignore
        name="Style",
        description="Default style for importing",
        items=STYLE_ITEMS,
        default="spheres",
    )
    import_md_topology: StringProperty(  # type: ignore
        name="Topology",
        description="File path for the toplogy file for the trajectory",
        subtype="FILE_PATH",
        maxlen=0,
    )
    import_md_trajectory: StringProperty(  # type: ignore
        name="Trajectory",
        description="File path for the trajectory file for the trajectory",
        subtype="FILE_PATH",
        maxlen=0,
    )
    import_md_name: StringProperty(  # type: ignore
        name="Name",
        description="Name of the molecule on import",
        default="NewTrajectory",
        maxlen=0,
    )
    import_density_invert: BoolProperty(  # type: ignore
        name="Invert Data",
        description="Invert the values in the map. Low becomes high, high becomes low.",
        default=False,
    )
    import_density_center: BoolProperty(  # type: ignore
        name="Center Density",
        description="Translate the density so that the center of the box is at the origin.",
        default=False,
    )
    import_density: StringProperty(  # type: ignore
        name="File",
        description="File path for the map file.",
        subtype="FILE_PATH",
        maxlen=0,
    )

    import_density_style: EnumProperty(  # type: ignore
        name="Style",
        items=(
            (
                "density_surface",
                "Surface",
                "A mesh surface based on the specified threshold",
                0,
            ),
            (
                "density_wire",
                "Wire",
                "A wire mesh surface based on the specified threshold",
                1,
            ),
        ),
    )

    panel_selection: bpy.props.EnumProperty(  # type: ignore
        name="Panel Selection",
        items=(
            ("import", "Import", "Import macromolecules", 0),
            ("object", "Object", "Adjust settings affecting the selected object", 1),
            (
                "session",
                "Session",
                "Interacting with the Molecular Nodes session tracking all of the objects",
                2,
            ),
        ),
    )

    panel_import_type: bpy.props.EnumProperty(  # type: ignore
        name="Method",
        items=(
            ("pdb", "PDB", "Download from the PDB"),
            ("alphafold", "AlphaFold", "Download from the AlphaFold DB"),
            ("local", "Local", "Open a local file"),
            ("md", "MD", "Import a molecular dynamics trajectory"),
            ("density", "Density", "Import an EM Density Map"),
            ("star", "Starfile", "Import a .starfile mapback file"),
            ("cellpack", "CellPack", "Import a CellPack .cif/.bcif file"),
            ("dna", "oxDNA", "Import an oxDNA file"),
        ),
    )
    import_star_file_path: StringProperty(  # type: ignore
        name="File",
        description="File path for the `.star` file to import.",
        subtype="FILE_PATH",
        maxlen=0,
    )
    import_cell_pack_path: bpy.props.StringProperty(  # type: ignore
        name="File",
        description="File to import (.cif, .bcif)",
        subtype="FILE_PATH",
        maxlen=0,
    )


class MolecularNodesObjectProperties(PropertyGroup):
    biological_assemblies: StringProperty(  # type: ignore
        name="Biological Assemblies",
        description="A list of biological assemblies to be created",
        default="",
    )

    entity_type: EnumProperty(  # type: ignore
        name="Entity Type",
        description="How the file was imported, dictating how MN interacts with it",
        items=(
            ("molecule", "Molecule", "A single molecule"),
            ("ensemble", "Ensemble", "A collection of molecules"),
            ("density", "Density", "An electron density map"),
            ("md", "Trajectory", "A molecular dynamics trajectory"),
            ("md-oxdna", "oxDNA Trajectory", "A oxDNA molecular dynamics trajectory "),
        ),
    )

    code: StringProperty(  # type: ignore
        name="PDB",
        description="PDB code used to download this structure",
        maxlen=4,
        options={"HIDDEN"},
    )
    trajectory_selection_index: IntProperty(  # type: ignore
        name="Index of selection",
        description="Index of selection, that is selected for the UI",
        default=0,
    )
    frame_hidden: IntProperty(  # type: ignore
        name="Frame",
        description="Frame of the loaded trajectory",
        default=0,
        min=0,
    )
    frame: IntProperty(  # type: ignore
        name="Frame",
        description="Frame of the loaded trajectory",
        default=0,
        update=_update_entities,
        min=0,
    )
    update_with_scene: BoolProperty(  # type: ignore
        name="Update with Scene",
        description="Update the trajectory with the scene frame",
        default=True,
        update=_update_entities,
    )
    subframes: IntProperty(  # type: ignore
        name="Subframes",
        description="Number of subframes to insert between frames of the loaded trajectory",
        default=0,
        update=_update_entities,
        min=0,
    )
    offset: IntProperty(  # type: ignore
        name="Offset",
        description="Offset the starting playback for the trajectory on the timeine. Positive starts the playback later than frame 0, negative starts it earlier than frame 0",
        default=0,
        update=_update_entities,
    )

    interpolate: BoolProperty(  # type: ignore
        name="Interpolate",
        description="Whether to interpolate when using subframes",
        default=True,
        update=_update_entities,
    )
    average: IntProperty(  # type: ignore
        name="Average",
        description="Average the position this number of frames either side of the current frame",
        default=0,
        update=_update_entities,
        min=0,
        soft_max=5,
    )
    correct_periodic: BoolProperty(  # type: ignore
        name="Correct",
        description="Correct for periodic boundary crossing when using interpolation or averaging. Assumes cubic dimensions and only works if the unit cell is orthorhombic",
        default=False,
        update=_update_entities,
    )
    filepath_trajectory: StringProperty(  # type: ignore
        name="Trajectory",
        description="Filepath for the `trajectory` of the Object",
        subtype="FILE_PATH",
        default="",
    )
    filepath_topology: StringProperty(  # type: ignore
        name="Topology",
        description="Filepath for the Topology of the Object",
        subtype="FILE_PATH",
        default="",
    )


class TrajectorySelectionItem(bpy.types.PropertyGroup):
    """Group of properties for custom selections for MDAnalysis import."""

    uuid: StringProperty(  # type: ignore
        name="UUID",
        description="Unique ID for matching selection in UI to selection on python object",
        default="",
    )

    name: StringProperty(  # type: ignore
        name="Name",
        description="Name of the attribute on the mesh",
        default="custom_selection",
        update=_update_entities,
    )

    selection_str: StringProperty(  # type: ignore
        name="Selection",
        description="Selection to be applied, written in the MDAnalysis selection language",
        default="name CA",
        update=_update_entities,
    )

    updating: BoolProperty(  # type: ignore
        name="Updating",
        description="Recalculate the selection on scene frame change",
        default=True,
        # update=_selection_update_trajectories,
    )

    periodic: BoolProperty(  # type: ignore
        name="Periodic",
        description="For geometric selections, whether to account for atoms in different periodic images when searching",
        default=True,
        # update=_selection_update_trajectories,
    )

    message: StringProperty(  # type: ignore
        name="Message",
        description="Message to report back from `universe.select_atoms()`",
        default="",
    )

    immutable: BoolProperty(  # type: ignore
        name="Immutable",
        description="Whether the selection is immutable",
        default=False,
    )


class MN_UL_TrajectorySelectionListUI(bpy.types.UIList):
    """UI List"""

    def draw_item(
        self,
        context,
        layout,
        data,
        item,
        icon,
        active_data,
        active_property,
        *,
        index=0,
        flt_flag=0,
    ):
        custom_icon = "VIS_SEL_11"

        if self.layout_type in {"DEFAULT", "COMPACT"}:
            row = layout.row()
            if item.message != "":
                custom_icon = "ERROR"
                row.alert = True

            col = row.column()
            col.prop(item, "name", text="", emboss=False)
            col.enabled = False
            row.prop(item, "updating", icon_only=True, icon="FILE_REFRESH")
            row.prop(item, "periodic", icon_only=True, icon="CUBE")
            if item.immutable:
                row.enabled = False

        elif self.layout_type in {"GRID"}:
            layout.alignment = "CENTER"
            layout.label(text="", icon=custom_icon)


class MN_OT_Universe_Selection_Add(bpy.types.Operator):
    "Add a new custom selection to a trajectory"

    bl_idname = "mn.trajectory_selection_add"
    bl_label = "+"
    bl_description = "Add a new boolean attribute for the given MDA selection string"

    def execute(self, context):
        obj = context.active_object
        traj = get_session(context).match(obj)
        i = int(len(obj.mn_trajectory_selections) - 1)
        name = "selection_0"
        while True:
            if len(obj.mn_trajectory_selections) == 0:
                break
            if name in obj.mn_trajectory_selections:
                i += 1
                name = f"selection_{i}"
            else:
                break
        traj.add_selection(name=name, selection_str="all")
        obj.mn["list_index"] = i

        return {"FINISHED"}


class MN_OT_Universe_Selection_Delete(bpy.types.Operator):
    bl_idname = "mda.delete_item"
    bl_label = "-"
    bl_description = "Delete the given boolean selection from the universe"

    @classmethod
    def poll(cls, context):
        return context.active_object.mn_trajectory_selections

    def execute(self, context):
        obj = context.active_object
        index = obj.mn.trajectory_selection_index
        traj = get_session(context).match(obj)
        names = [s.name for s in obj.mn_trajectory_selections]
        traj.remove_selection(names[index])
        obj.mn.trajectory_selection_index = int(
            max(min(index, len(obj.mn_trajectory_selections) - 1), 0)
        )

        return {"FINISHED"}


CLASSES = [
    MolecularNodesObjectProperties,
    MolecularNodesSceneProperties,
    TrajectorySelectionItem,  # item has to be registered the ListUI and to work properly
    MN_UL_TrajectorySelectionListUI,
    MN_OT_Universe_Selection_Add,
    MN_OT_Universe_Selection_Delete,
]



import bpy
from ..entities import Molecule
import tempfile
from IPython.display import Image, display
from pathlib import Path

IS_EEVEE_NEXT = bpy.app.version[0] == 4 and bpy.app.version[1] >= 2


class Canvas:
    """
    A class to handle canvas and render settings in Blender.

    This class provides properties to get and set various render settings like resolution,
    render engine, samples, FPS, and frame ranges.
    """

    def __init__(self) -> None:
        """Initialize the Canvas object."""
        pass

    @property
    def scene(self) -> bpy.types.Scene:
        return bpy.context.scene

    @property
    def resolution(self) -> tuple[int, int]:
        """
        Get the render resolution.

        Returns
        -------
        tuple[int, int]
            A tuple containing the x and y resolution values.
        """
        x: int = self.scene.render.resolution_x
        y: int = self.scene.render.resolution_y
        return (x, y)

    @resolution.setter
    def resolution(self, value: tuple[int, int]) -> None:
        """
        Set the render resolution.

        Parameters
        ----------
        value : tuple[int, int]
            A tuple containing the x and y resolution values.
        """
        self.scene.render.resolution_x = value[0]
        self.scene.render.resolution_y = value[1]

    @property
    def render_engine(self) -> str:
        """
        The current render engine for the scene.

        Returns
        -------
        str
            The name of the current render engine.
            Possible values are:
            - 'BLENDER_EEVEE': Eevee real-time render engine
            - 'CYCLES': Cycles path tracing render engine
            - 'BLENDER_WORKBENCH': Workbench render engine
        """
        return self.scene.render.engine

    @render_engine.setter
    def render_engine(self, value: str) -> None:
        """
        Set the render engine.

        Parameters
        ----------
        value : str
            The name of the render engine to set.
            Possible values are:
            - 'BLENDER_EEVEE': Eevee real-time render engine
            - 'CYCLES': Cycles path tracing render engine
            - 'BLENDER_WORKBENCH': Workbench render engine
        """

        if value == "EEVEE":
            if IS_EEVEE_NEXT:
                value = "BLENDER_EEVEE_NEXT"
            else:
                value = "BLENDER_EEVEE"

        if value == "WORKBENCH":
            value = "BLENDER_WORKBENCH"
        self.scene.render.engine = value  # type: ignore

    @property
    def samples_cycles(self) -> int:
        """
        Get the number of samples for Cycles render engine.

        Returns
        -------
        int
            The number of samples for Cycles.
        """
        return self.scene.cycles.samples

    @samples_cycles.setter
    def samples_cycles(self, value: int) -> None:
        """
        Set the number of samples for Cycles render engine.

        Parameters
        ----------
        value : int
            The number of samples to set for Cycles.
        """
        self.scene.cycles.samples = value

    @property
    def samples_eevee(self) -> int:
        """
        Get the number of samples for Eevee render engine.

        Returns
        -------
        int
            The number of samples for Eevee.
        """
        return self.scene.eevee.taa_render_samples

    @samples_eevee.setter
    def samples_eevee(self, value: int) -> None:
        """
        Set the number of samples for Eevee render engine.

        Parameters
        ----------
        value : int
            The number of samples to set for Eevee.
        """
        self.scene.eevee.taa_render_samples = value

    @property
    def cycles_device(self) -> str:
        """
        Get the Cycles device type.
        """
        return self.scene.cycles.device

    @cycles_device.setter
    def cycles_device(self, value: str) -> None:
        """
        Set the Cycles device type.
        """
        value = value.upper()
        if value not in ["CPU", "GPU"]:
            raise ValueError("Invalid device type. Must be 'CPU' or 'GPU'.")
        self.scene.cycles.device = value

    @property
    def fps(self) -> float:
        """
        Get the frames per second setting.

        Returns
        -------
        float
            The current FPS value.
        """
        return self.scene.render.fps

    @fps.setter
    def fps(self, value: float) -> None:
        """
        Set the frames per second.

        Parameters
        ----------
        value : float
            The FPS value to set.
        """
        self.scene.render.fps = value

    @property
    def frame_start(self) -> int:
        """
        Get the start frame of the animation.

        Returns
        -------
        int
            The start frame number.
        """
        return self.scene.frame_start

    @frame_start.setter
    def frame_start(self, value: int) -> None:
        """
        Set the start frame of the animation.

        Parameters
        ----------
        value : int
            The start frame number to set.
        """
        self.scene.frame_start = value

    @property
    def frame_end(self) -> int:
        """
        Get the end frame of the animation.

        Returns
        -------
        int
            The end frame number.
        """
        return self.scene.frame_end

    @frame_end.setter
    def frame_end(self, value: int) -> None:
        """
        Set the end frame of the animation.

        Parameters
        ----------
        value : int
            The end frame number to set.
        """
        self.scene.frame_end = value

    def frame_object(self, obj: bpy.types.Object | Molecule) -> None:
        if isinstance(obj, Molecule):
            obj = obj.object

        prev_sel = bpy.context.selected_objects
        obj.select_set(True)
        bpy.ops.view3d.camera_to_view_selected()
        obj.select_set(False)
        for o in prev_sel:
            o.select_set(True)

    def scene_reset(self) -> None:
        bpy.ops.wm.read_homefile(app_template="Molecular Nodes")

    def snapshot(self, path: str | Path | None = None) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir) / "snapshot.png"
            bpy.context.scene.render.filepath = str(tmp_path)
            bpy.ops.render.render(write_still=True, animation=False)
            display(Image(tmp_path))



import bpy

from ..entities.trajectory import dna

from ..blender import nodes
from ..session import get_session
from ..entities import density, ensemble, molecule, trajectory


chosen_panel = {
    "pdb": molecule.ui.panel_wwpdb,
    "local": molecule.ui.panel_local,
    "alphafold": molecule.ui.panel_alphafold,
    "star": ensemble.ui.panel_starfile,
    "md": trajectory.ui.panel,
    "density": density.ui.panel,
    "cellpack": ensemble.ui.panel_cellpack,
    "dna": dna.panel,
}


def pt_object_context(self, context):
    layout = self.layout
    return None


def is_style_node(context):
    node = context.space_data.edit_tree.nodes.active
    return node.name.startswith("Style")


def change_style_node_menu(self, context):
    layout = self.layout
    node = context.active_node

    # return early if not a node group
    if not hasattr(node, "node_tree"):
        return None

    # return early if the node group isn't one of the ones we want to swap easily
    prefix = node.node_tree.name.split(" ")[0].lower()
    if prefix not in ["color", "select", "is", "style", "topology", "animate"]:
        return None

    layout.label(text="Molecular Nodes", icon="MOD_PARTICLES")

    row = layout.row()
    op = row.operator_menu_enum("mn.node_swap", "node_items", text="Change Node")
    op.node_description = "The topology nodes"

    layout.separator()


def panel_import(layout, context):
    scene = context.scene
    selection = scene.mn.panel_import_type
    layout.prop(scene.mn, "panel_import_type")

    col = layout.column()
    chosen_panel[selection](col, scene)


def ui_from_node(
    layout: bpy.types.UILayout, node: bpy.types.NodeGroup, context: bpy.types.Context
):
    """
    Generate the UI for a particular node, which displays the relevant node inputs
    for user control in a panel, rather than through the node editor.
    """
    col = layout.column(align=True)
    ntree = context.active_object.modifiers["MolecularNodes"].node_group

    tree = node.node_tree.interface.items_tree

    for item in tree.values():
        if item.item_type == "PANEL":
            col.label(text=item.name)
        elif item.name == "Selection":
            continue
        else:
            if item.in_out != "INPUT":
                continue
            if item.socket_type == "NodeSocketGeometry":
                continue
            col.template_node_view(ntree, node, node.inputs[item.identifier])


def panel_md_properties(layout, context):
    obj = context.active_object
    session = get_session()
    traj: trajectory.Trajectory = session.match(obj)
    traj_is_linked = bool(traj)
    if traj is not None and not isinstance(traj, trajectory.Trajectory):
        raise TypeError(f"Expected a trajectory, got {type(traj)}")

    col = layout.column()
    col.enabled = False
    if not traj_is_linked:
        col.enabled = True
        col.label(text="Object not linked to a trajectory, please reload one")
        col.prop(obj.mn, "filepath_topology")
        col.prop(obj.mn, "filepath_trajectory")
        col.operator("mn.reload_trajectory")
        return None

    layout.label(text="Trajectory Playback", icon="OPTIONS")

    row = layout.row()
    col = row.column()
    if obj.mn.update_with_scene:
        col.prop(obj.mn, "frame_hidden")
    else:
        col.prop(obj.mn, "frame")
    col.enabled = not obj.mn.update_with_scene
    row.prop(obj.mn, "update_with_scene")
    row = layout.row()
    col = row.column()
    col.enabled = obj.mn.update_with_scene
    col.prop(obj.mn, "average")
    col.prop(obj.mn, "subframes")
    col.prop(obj.mn, "offset")
    col = row.column()
    col.enabled = obj.mn.update_with_scene

    # only enable this as an option if the universe is orthothombic
    row = col.row()
    row.prop(obj.mn, "correct_periodic")
    row.enabled = traj.is_orthorhombic
    col.prop(obj.mn, "interpolate")

    layout.label(text="Selections", icon="RESTRICT_SELECT_OFF")
    row = layout.row()
    row = row.split(factor=0.9)
    row.template_list(
        "MN_UL_TrajectorySelectionListUI",
        "A list",
        obj,
        "mn_trajectory_selections",
        obj.mn,
        "trajectory_selection_index",
        rows=3,
    )
    col = row.column()
    col.operator("mn.trajectory_selection_add", icon="ADD", text="")
    col.operator("mda.delete_item", icon="REMOVE", text="")
    if obj.mn_trajectory_selections:
        item = obj.mn_trajectory_selections[obj.mn.trajectory_selection_index]

        col = layout.column(align=False)
        row = col.row()
        col.prop(item, "selection_str")

        # disable editing for immutable selections
        # disable modifying updating and periodic
        if item.immutable:
            col.enabled = False

        if item.message != "":
            box = col.box()
            box.label(text="Invalid Selection", icon="ERROR")
            box.label(text=item.message)
            box.alert = True
            op = box.operator("wm.url_open", text="Selection Langauge Docs", icon="URL")
            op.url = (
                "https://docs.mdanalysis.org/stable/documentation_pages/selections.html"
            )


def panel_object(layout, context):
    object = context.active_object
    layout.prop(object.mn, "entity_type")
    try:
        mol_type = object.mn.entity_type
    except AttributeError:
        return None
    if mol_type == "":
        layout.label(text="No MN object selected")
        return None
    if mol_type == "pdb":
        layout.label(text=f"PDB: {object.mn.code.upper()}")
    if mol_type.startswith("md"):
        panel_md_properties(layout, context)
    if mol_type == "star":
        layout.label(text="Ensemble")
        box = layout.box()
        ui_from_node(box, nodes.get_star_node(object), context=context)
        return None


def item_ui(layout, item):
    row = layout.row()
    row.label(text=item.name)
    col = row.column()
    op = col.operator("mn.session_create_object")
    op.uuid = item.uuid
    col.enabled = item.object is None

    op = row.operator("mn.session_remove_item", text="", icon="CANCEL")
    op.uuid = item.uuid

    if item.object is not None:
        row = layout.row()
        row.label(text=f"Object: {item.object.name}", icon="OUTLINER_OB_MESH")


def panel_session(layout, context):
    session = get_session(context)
    # if session.n_items > 0:
    #     return None
    row = layout.row()
    row.label(text="Loaded items in the session")
    # row.operator("mn.session_reload")

    layout.label(text="Molecules")
    box = layout.box()
    for mol in session.molecules.values():
        item_ui(box, mol)

    layout.label(text="Universes")
    box = layout.box()
    for uni in session.trajectories.values():
        item_ui(box, uni)

    layout.label(text="Ensembles")
    box = layout.box()
    for ens in session.ensembles.values():
        item_ui(box, ens)


def panel_scene(layout, context):
    scene = context.scene

    cam = bpy.data.cameras[bpy.data.scenes["Scene"].camera.name]
    world_shader = bpy.data.worlds["World Shader"].node_tree.nodes["MN_world_shader"]
    grid = layout.grid_flow()
    col = grid.column()
    col.label(text="World Settings")
    world = col.box()
    world.prop(bpy.data.scenes["Scene"].render, "engine")
    if scene.render.engine == "CYCLES":
        world.prop(bpy.data.scenes["Scene"].cycles, "samples")
    else:
        world.prop(bpy.data.scenes["Scene"].eevee, "taa_render_samples")
    world.label(text="Background")
    world.prop(world_shader.inputs[1], "default_value", text="HDRI Strength")
    row = world.row()
    row.prop(scene.render, "film_transparent")
    row.prop(world_shader.inputs[2], "default_value", text="Background")

    col = grid.column()
    col.label(text="Camera Settings")
    camera = col.box()
    camera.prop(cam, "lens")
    col = camera.column(align=True)
    row = col.row(align=True)
    row.prop(bpy.data.scenes["Scene"].render, "resolution_x", text="X")
    row.prop(bpy.data.scenes["Scene"].render, "resolution_y", text="Y")
    row = camera.grid_flow()
    row.prop(cam.dof, "use_dof")
    row.prop(bpy.data.scenes["Scene"].render, "use_motion_blur")
    focus = camera.column()
    focus.enabled = cam.dof.use_dof
    focus.prop(cam.dof, "focus_object")
    distance = focus.row()
    distance.enabled = cam.dof.focus_object is None
    distance.prop(cam.dof, "focus_distance")
    focus.prop(cam.dof, "aperture_fstop")


class MN_PT_Scene(bpy.types.Panel):
    bl_label = "Molecular Nodes"
    bl_idname = "MN_PT_panel"
    bl_space_type = "PROPERTIES"
    bl_region_type = "WINDOW"
    bl_context = "scene"
    bl_order = 0
    bl_options = {"HEADER_LAYOUT_EXPAND"}
    bl_ui_units_x = 0

    def draw(self, context):
        layout = self.layout
        scene = context.scene
        row = layout.row()
        # row.operator("mn.import_fetch").database = "wwpdb"
        # row.operator("mn.import_local")
        # row.operator("mn.import_fetch").database = "alphafold"
        # row.operator("mn.import_trajectory")

        row = layout.row(align=True)

        for p in ["import", "object", "session"]:
            row.prop_enum(scene.mn, "panel_selection", p)

        # the possible panel functions to choose between
        which_panel = {
            "import": panel_import,
            "object": panel_object,
            "session": panel_session,
        }
        # call the required panel function with the layout and context
        which_panel[scene.mn.panel_selection](layout, context)


CLASSES = []



import numpy as np

import molecularnodes as mn
from molecularnodes.blender import nodes
import databpy

from .constants import data_dir
import pytest


def test_centre_on_selection():
    mol = mn.entities.fetch("1cd3", cache_dir=data_dir, style=None)

    group = nodes.get_mod(mol.object).node_group = nodes.new_tree()
    link = group.links.new
    node_centre = nodes.add_custom(group, "Centre on Selection")
    node_chain_id = nodes.add_custom(group, "Chain ID", location=[-100, -100])

    link(group.nodes["Group Input"].outputs[0], node_centre.inputs[0])
    link(node_chain_id.outputs["chain_id"], node_centre.inputs["Group ID"])
    link(node_centre.outputs[0], group.nodes["Group Output"].inputs[0])

    old_pos = mol.named_attribute("position", evaluate=False)
    new_pos = mol.named_attribute("position", evaluate=True)

    assert not np.allclose(old_pos, new_pos)

    chain_id = mol.named_attribute("chain_id")
    chain_ids = np.unique(chain_id)

    old_centres = [databpy.centre(old_pos[chain_id == x]) for x in chain_ids]
    new_centres = [databpy.centre(new_pos[chain_id == x]) for x in chain_ids]

    assert not np.allclose(old_centres, new_centres)
    assert np.allclose(
        new_centres, [np.array((0, 0, 0), dtype=float) for x in chain_ids], atol=1e-5
    )


def test_atoms_to_ca_splines(snapshot):
    """
    This test is for if the chains are still connected as continuous, even though they are
    separate by a larger distance
    """
    mol = mn.entities.fetch("1HQM", style=None)

    group = nodes.get_mod(mol.object).node_group = nodes.new_tree()
    link = group.links.new

    node_atcas = nodes.add_custom(group, ".Atoms to CA Splines")
    node_ctm = group.nodes.new("GeometryNodeCurveToMesh")

    link(group.nodes["Group Input"].outputs[0], node_atcas.inputs[0])
    link(node_atcas.outputs[0], node_ctm.inputs[0])
    link(node_ctm.outputs[0], group.nodes["Group Output"].inputs[0])

    pos_pre = mol.named_attribute("position", evaluate=True)

    node_atcas.inputs["Threshold (A)"].default_value = 200

    pos_post = mol.named_attribute("position", evaluate=True)

    assert not np.allclose(pos_pre.shape, pos_post.shape)



import bpy
import molecularnodes as mn
import pytest

from .constants import data_dir


def test_session_present():
    assert isinstance(mn.session.get_session(), mn.session.MNSession)


def test_persistent_handlers_added():
    load_handlers = [handler.__name__ for handler in bpy.app.handlers.load_post]
    save_handlers = [handler.__name__ for handler in bpy.app.handlers.save_post]
    assert "_pickle" in save_handlers
    assert "_load" in load_handlers


def test_entity_registered():
    session = mn.session.get_session()
    assert len(session.entities) == 0

    mol = mn.entities.fetch("1BNA", cache_dir=data_dir)

    assert mol.uuid in session.entities
    assert isinstance(session.get(mol.uuid), mn.entities.BCIF)
    assert len(session.entities) == 1



from .base import Molecule
from biotite.structure.io.mol import MOLFile


class SDF(Molecule):
    def __init__(self, file_path):
        super().__init__(file_path)
        self.file = self.read(self.file_path)
        self.array = self._get_structure()
        self.n_atoms = self.array.array_length()

    def read(self, file_path):
        return MOLFile.read(file_path)

    def _get_structure(self):
        return self.file.get_structure()

    def _assemblies(self):
        # TODO maybe look into symmetry operations for small mols
        return None



import itertools

import biotite.structure as struc
import biotite.structure.io.pdbx as pdbx
import numpy as np
from biotite import InvalidFileError

from .base import Molecule


class PDBX(Molecule):
    def __init__(self, file_path):
        super().__init__(file_path=file_path)
        self._extra_annotations = {
            "sec_struct": self._get_secondary_structure,
            "entity_id": self._get_entity_id,
        }

    @property
    def entity_ids(self):
        return self.file.block.get("entity").get("pdbx_description").as_array().tolist()

    def set_extra_annotations(
        self,
        array: struc.AtomArray,
        file: pdbx.BinaryCIFFile | pdbx.CIFFile,
        verbose: bool = False,
    ) -> struc.AtomArray:
        """Set new annotations on the array from custom functions.

        The custom functions are stored in the self._extra_annotations dictionary. They
        take the array and the file as arguments and return the new annotation as a numpy
        array for passing into the `AtomArray.set_annotation(name, array)` function.

        Parameters
        ----------
        array : struc.AtomArray
            The atom array to add annotations to
        file : pdbx.BinaryCIFFile | pdbx.CIFFile
            The CIF file containing the annotation data
        verbose : bool, optional
            Whether to print error messages, by default False

        Returns
        -------
        struc.AtomArray
            The atom array with added annotations
        """

        for name, func in self._extra_annotations.items():
            try:
                # for the getting of some custom attributes, we get it for the full atom_site
                # but we need to assign a subset of the array that is that model in the full
                # atom_site, so we subset using the atom_id
                try:
                    array.set_annotation(name, func(array, file))
                except IndexError:
                    array.set_annotation(name, func(array, file)[array.atom_id - 1])
            except KeyError as e:
                if verbose:
                    print(f"Unable to add {name} as an attribute, error: {e}")

        return array

    def get_structure(
        self,
        extra_fields=["b_factor", "occupancy", "atom_id"],
        bonds=True,
        model: int | None = None,
    ):
        try:
            array = pdbx.get_structure(
                self.file, model=model, extra_fields=extra_fields
            )
            array = self.set_extra_annotations(array, self.file)
            if not array.bonds and bonds:
                array.bonds = struc.bonds.connect_via_residue_names(
                    array, inter_residue=True
                )
        except InvalidFileError:
            array = pdbx.get_component(self.file)
            if not array.bonds and bonds:
                array.bonds = struc.bonds.connect_via_residue_names(
                    array, inter_residue=True
                )

        return array

    def _assemblies(self):
        return CIFAssemblyParser(self.file).get_assemblies()

    @staticmethod
    def _extract_matrices(category):
        matrix_columns = [
            "matrix[1][1]",
            "matrix[1][2]",
            "matrix[1][3]",
            "vector[1]",
            "matrix[2][1]",
            "matrix[2][2]",
            "matrix[2][3]",
            "vector[2]",
            "matrix[3][1]",
            "matrix[3][2]",
            "matrix[3][3]",
            "vector[3]",
        ]

        columns = [category[name].as_array().astype(float) for name in matrix_columns]
        matrices = np.empty((len(columns[0]), 4, 4), float)

        col_mask = np.tile((0, 1, 2, 3), 3)
        row_mask = np.repeat((0, 1, 2), 4)
        for column, coli, rowi in zip(columns, col_mask, row_mask):
            matrices[:, rowi, coli] = column

        return matrices

    @staticmethod
    def _get_entity_id(array, file):
        chain_ids = file.block["entity_poly"]["pdbx_strand_id"].as_array(str)

        # the chain_ids are an array of individual items np.array(['A,B', 'C', 'D,E,F'])
        # which need to be categorised as [1, 1, 2, 3, 3, 3] for their belonging to individual
        # entities

        chains = []
        idx = []
        for i, chain_str in enumerate(chain_ids):
            for chain in chain_str.split(","):
                chains.append(chain)
                idx.append(i)

        # this is how we map the chain_ids and res_names of our entities to their integer
        # representations
        entity_lookup = dict(zip(chains, idx))

        # for the hetero atoms, we need to add a new entity_id into the lookup so that
        # they can be assigned an entity ID
        unique_res_het = np.unique(array.res_name[array.hetero])
        for het in unique_res_het:
            if het not in entity_lookup:
                entity_lookup[het] = max(entity_lookup.values()) + 1

        entity_id_int = np.zeros(len(array.res_name), int)
        for i, res_name in enumerate(array.res_name):
            entity_id_int[i] = entity_lookup.get(
                res_name, entity_lookup[array.chain_id[i]]
            )

        return entity_id_int

    @staticmethod
    def _get_secondary_structure(array, file):
        """
        Get secondary structure information for the array from the file.

        Parameters
        ----------
        array : numpy array
            The array for which secondary structure information is to be retrieved.
        file : object
            The file object containing the secondary structure information.

        Returns
        -------
        numpy array
            A numpy array of secondary structure information, where each element is either 0, 1, 2, or 3.
            - 0: Not a peptide
            - 1: Alpha helix
            - 2: Beta sheet
            - 3: Loop

        Raises
        ------
        KeyError
            If the 'struct_conf' category is not found in the file.
        """

        # get the annotations for the struc_conf cetegory. Provides start and end
        # residues for the annotations. For most files this will only contain the
        # alpha helices, but will sometimes contain also other secondary structure
        # information such as in AlphaFold predictions

        conf = file.block.get("struct_conf")
        if conf:
            starts = conf["beg_auth_seq_id"].as_array().astype(int)
            ends = conf["end_auth_seq_id"].as_array().astype(int)
            chains = conf["end_auth_asym_id"].as_array().astype(str)
            id_label = conf["id"].as_array().astype(str)
        else:
            starts = np.empty(0, dtype=int)
            ends = np.empty(0, dtype=int)
            chains = np.empty(0, dtype=str)
            id_label = np.empty(0, dtype=int)

        # most files will have a separate category for the beta sheets
        # this can just be appended to the other start / end / id and be processed
        # as normalquit
        sheet = file.block.get("struct_sheet_range")
        if sheet:
            starts = np.append(starts, sheet["beg_auth_seq_id"].as_array().astype(int))
            ends = np.append(ends, sheet["end_auth_seq_id"].as_array().astype(int))
            chains = np.append(chains, sheet["end_auth_asym_id"].as_array().astype(str))
            id_label = np.append(id_label, np.repeat("STRN", len(sheet["id"])))

        if not conf and not sheet:
            raise KeyError

        # convert the string labels to integer representations of the SS
        # AH: 1, BS: 2, LOOP: 3

        id_int = np.array([_ss_label_to_int(label) for label in id_label], int)

        # create a lookup dictionary that enables lookup of secondary structure
        # based on the chain_id and res_id values

        lookup = dict()
        for chain in np.unique(chains):
            arrays = []
            mask = chain == chains
            start_sub = starts[mask]
            end_sub = ends[mask]
            id_sub = id_int[mask]

            for start, end, id in zip(start_sub, end_sub, id_sub):
                idx = np.arange(start, end + 1, dtype=int)
                arr = np.zeros((len(idx), 2), dtype=int)
                arr[:, 0] = idx
                arr[:, 1] = 3
                arr[:, 1] = id
                arrays.append(arr)

            lookup[chain] = dict(np.vstack(arrays).tolist())

        # use the lookup dictionary to get the SS annotation based on the chain_id and res_id
        secondary_structure = np.zeros(len(array.chain_id), int)
        for i, (chain, res) in enumerate(zip(array.chain_id, array.res_id)):
            try:
                secondary_structure[i] = lookup[chain].get(res, 3)
            except KeyError:
                secondary_structure[i] = 0

        # assign SS to 0 where not peptide
        secondary_structure[~struc.filter_amino_acids(array)] = 0
        return secondary_structure


def _parse_opers(oper):
    # we want the example '1,3,(5-8)' to expand to (1, 3, 5, 6, 7, 8).
    op_ids = list()

    for group in oper.strip(")").split("("):
        if "," in group:
            for i in group.split(","):
                op_ids.append()

    for group in oper.split(","):
        if "-" not in group:
            op_ids.append(str(group))
            continue

        start, stop = [int(x) for x in group.strip("()").split("-")]
        for i in range(start, stop + 1):
            op_ids.append(str(i))

    return op_ids


def _ss_label_to_int(label):
    if "HELX" in label:
        return 1
    elif "STRN" in label:
        return 2
    else:
        return 3


class CIF(PDBX):
    def __init__(self, file_path):
        super().__init__(file_path)
        # self.file_path = file_path
        # self.file = self.read(file_path)
        self.array = self.get_structure()

    def _read(self, file_path):
        return pdbx.CIFFile.read(file_path)


class BCIF(PDBX):
    def __init__(self, file_path):
        super().__init__(file_path)
        # self.file_path = file_path
        # self.file = self.read(file_path)
        self.array = self.get_structure()

    def _read(self, file_path):
        return pdbx.BinaryCIFFile.read(file_path)


class CIFAssemblyParser:
    # Implementation adapted from ``biotite.structure.io.pdbx.convert``

    def __init__(self, file_cif):
        self._file = file_cif

    def list_assemblies(self):
        return list(pdbx.list_assemblies(self._file).keys())

    def get_transformations(self, assembly_id):
        assembly_gen_category = self._file.block["pdbx_struct_assembly_gen"]

        struct_oper_category = self._file.block["pdbx_struct_oper_list"]

        if assembly_id not in assembly_gen_category["assembly_id"].as_array(str):
            raise KeyError(f"File has no Assembly ID '{assembly_id}'")

        # Extract all possible transformations indexed by operation ID
        # transformation_dict = _get_transformations(struct_oper_category)
        transformation_dict = _extract_matrices(struct_oper_category)

        # Get necessary transformations and the affected chain IDs
        # NOTE: The chains given here refer to the `label_asym_id` field
        # of the `atom_site` category
        # However, by default `PDBxFile` uses the `auth_asym_id` as
        # chain ID
        matrices = []
        pdb_model_num = -1
        for id, op_expr, asym_id_expr in zip(
            assembly_gen_category["assembly_id"].as_array(str),
            assembly_gen_category["oper_expression"].as_array(str),
            assembly_gen_category["asym_id_list"].as_array(str),
        ):
            pdb_model_num += 1
            # Find the operation expressions for given assembly ID
            # We already asserted that the ID is actually present
            if id != assembly_id:
                continue

            operations = _parse_operation_expression(op_expr)

            affected_chain_ids = asym_id_expr.split(",")

            for i, operation in enumerate(operations):
                # for op_step in operation:
                matrices.append(
                    {
                        "chain_ids": affected_chain_ids,
                        "matrix": transformation_dict[operation[0]].tolist(),
                        "pdb_model_num": pdb_model_num,
                    }
                )

        return matrices

    def get_assemblies(self):
        assembly_dict = {}
        for assembly_id in self.list_assemblies():
            assembly_dict[assembly_id] = self.get_transformations(assembly_id)

        return assembly_dict


def _extract_matrices(category, scale=True):
    matrix_columns = [
        "matrix[1][1]",
        "matrix[1][2]",
        "matrix[1][3]",
        "vector[1]",
        "matrix[2][1]",
        "matrix[2][2]",
        "matrix[2][3]",
        "vector[2]",
        "matrix[3][1]",
        "matrix[3][2]",
        "matrix[3][3]",
        "vector[3]",
    ]

    columns = [category[name].as_array().astype(float) for name in matrix_columns]
    n = 4 if scale else 3
    matrices = np.empty((len(columns[0]), n, 4), float)

    col_mask = np.tile((0, 1, 2, 3), 3)
    row_mask = np.repeat((0, 1, 2), 4)
    for column, coli, rowi in zip(columns, col_mask, row_mask):
        matrices[:, rowi, coli] = column

    return dict(zip(category["id"].as_array(str), matrices))


def _chain_transformations(rotations, translations):
    """
    Get a total rotation/translation transformation by combining
    multiple rotation/translation transformations.
    This is done by intermediately combining rotation matrices and
    translation vectors into 4x4 matrices in the form

    |r11 r12 r13 t1|
    |r21 r22 r23 t2|
    |r31 r32 r33 t3|
    |0   0   0   1 |.
    """
    total_matrix = np.identity(4)
    for rotation, translation in zip(rotations, translations):
        matrix = np.zeros((4, 4))
        matrix[:3, :3] = rotation
        matrix[:3, 3] = translation
        matrix[3, 3] = 1
        total_matrix = matrix @ total_matrix

    # return total_matrix[:3, :3], total_matrix[:3, 3]
    return matrix


def _get_transformations(struct_oper):
    """
    Get transformation operation in terms of rotation matrix and
    translation for each operation ID in ``pdbx_struct_oper_list``.
    """
    transformation_dict = {}
    for index, id in enumerate(struct_oper["id"].as_array()):
        rotation_matrix = np.array(
            [
                [float(struct_oper[f"matrix[{i}][{j}]"][index]) for j in (1, 2, 3)]
                for i in (1, 2, 3)
            ]
        )
        translation_vector = np.array(
            [float(struct_oper[f"vector[{i}]"][index]) for i in (1, 2, 3)]
        )
        transformation_dict[id] = (rotation_matrix, translation_vector)
    return transformation_dict


def _parse_operation_expression(expression):
    """
    Get successive operation steps (IDs) for the given
    ``oper_expression``.
    Form the cartesian product, if necessary.
    """
    # Split groups by parentheses:
    # use the opening parenthesis as delimiter
    # and just remove the closing parenthesis
    expressions_per_step = expression.replace(")", "").split("(")
    expressions_per_step = [e for e in expressions_per_step if len(e) > 0]
    # Important: Operations are applied from right to left
    expressions_per_step.reverse()

    operations = []
    for expr in expressions_per_step:
        if "-" in expr:
            if "," in expr:
                for gexpr in expr.split(","):
                    if "-" in gexpr:
                        first, last = gexpr.split("-")
                        operations.append(
                            [str(id) for id in range(int(first), int(last) + 1)]
                        )
                    else:
                        operations.append([gexpr])
            else:
                # Range of operation IDs, they must be integers
                first, last = expr.split("-")
                operations.append([str(id) for id in range(int(first), int(last) + 1)])
        elif "," in expr:
            # List of operation IDs
            operations.append(expr.split(","))
        else:
            # Single operation ID
            operations.append([expr])

    # Cartesian product of operations
    return list(itertools.product(*operations))



import molecularnodes as mn
import numpy as np


def test_correct_1d():
    assert np.allclose(
        mn.utils.correct_periodic_1d(np.array((0.9, 0.1)), np.array((0.1, 0.9)), 1.0),
        np.array((1.1, -0.1)),
    )



import molecularnodes as mn
import pytest
from scipy.spatial.transform import Rotation as R
import starfile
from .constants import data_dir


@pytest.mark.parametrize("type", ["cistem", "relion"])
def test_starfile_attributes(type):
    file = data_dir / f"starfile/{type}.star"
    ensemble = mn.entities.ensemble.load_starfile(file)

    star = starfile.read(file)

    if type == "relion":
        df = star["particles"].merge(star["optics"], on="rlnOpticsGroup")
        euler_angles = df[["rlnAngleRot", "rlnAngleTilt", "rlnAnglePsi"]].to_numpy()

    elif type == "cistem":
        df = star
        euler_angles = df[
            ["cisTEMAnglePhi", "cisTEMAngleTheta", "cisTEMAnglePsi"]
        ].to_numpy()

    # Calculate Scipy rotation from the euler angles
    rot_from_euler = quats = R.from_euler(
        seq="ZYZ", angles=euler_angles, degrees=True
    ).inv()

    # Convert from blender to scipy conventions and then into Scipy rotation
    quat_attribute = ensemble.named_attribute("rotation")
    rot_from_geo_nodes = R.from_quat(quat_attribute[:, [1, 2, 3, 0]])

    # To compare the two rotation with multiply one with the inverse of the other
    assert (rot_from_euler * rot_from_geo_nodes.inv()).magnitude().max() < 1e-5


def test_load_starfiles():
    file = data_dir / "starfile/clathrin.star"
    ensemble = mn.entities.ensemble.load_starfile(file)


def test_categorical_attributes():
    file = data_dir / "starfile/cistem.star"
    ensemble = mn.entities.ensemble.load_starfile(file)
    assert "cisTEMOriginalImageFilename_categories" in ensemble.object


def test_micrograph_conversion():
    file = data_dir / "starfile/cistem.star"
    ensemble = mn.entities.ensemble.load_starfile(file)
    tiff_path = data_dir / "starfile/montage.tiff"
    tiff_path.unlink(missing_ok=True)
    ensemble._convert_mrc_to_tiff()
    assert tiff_path.exists()


def test_micrograph_loading():
    import bpy

    file = data_dir / "starfile/cistem.star"
    tiff_path = data_dir / "starfile/montage.tiff"
    tiff_path.unlink(missing_ok=True)

    ensemble = mn.entities.ensemble.load_starfile(file)
    assert not tiff_path.exists()
    ensemble.star_node.inputs["Show Micrograph"].default_value = True
    bpy.context.evaluated_depsgraph_get().update()
    assert tiff_path.exists()
    # Ensure montage get only loaded once
    assert sum(1 for image in bpy.data.images.keys() if "montage" in image) == 1
    assert (
        ensemble.micrograph_material.node_tree.nodes["Image Texture"].image.name
        == "montage.tiff"
    )
    assert ensemble.star_node.inputs["Micrograph"].default_value.name == "montage.tiff"



import itertools

import numpy as np
import pytest

import molecularnodes as mn

from .constants import attributes, codes, data_dir


formats = ["pdb", "cif", "bcif"]


@pytest.mark.parametrize("code, format", itertools.product(codes, formats))
def test_attribute(snapshot_custom, code, format):
    mol = mn.entities.fetch(code, cache_dir=data_dir, style=None, format=format)
    for attribute in attributes:
        try:
            assert snapshot_custom == mol.named_attribute(attribute)
        except AttributeError as e:
            assert snapshot_custom == e


def test_store_named_attribute(snapshot_custom):
    mol = mn.entities.fetch("8H1B", cache_dir=data_dir, style=None, format="bcif")
    before = mol.named_attribute("position")
    mol.store_named_attribute(mol.named_attribute("position") + 10, "position")
    after = mol.named_attribute("position")

    assert not np.allclose(before, after)



import MDAnalysis as mda
import numpy.typing as npt
import numpy as np
from uuid import uuid1

class Selection:
    def __init__(
        self, trajectory, name: str = "selection_0", selection_str: str = "all",  updating=True, periodic=True
    ):  
        self._ag: mda.AtomGroup | None = None
        self._uuid: str = str(uuid1())
        self._name = name
        self._current_selection_str: str = ""
        self.trajectory = trajectory
        self.trajectory.object.mn_trajectory_selections.add()
        self.trajectory.object.mn_trajectory_selections[-1].name = self.name
        self.trajectory.object.mn_trajectory_selections[-1].uuid = self._uuid
        self.updating = updating
        self.periodic = periodic
        self.set_atom_group(selection_str)
        self.selection_str = selection_str
        self.mask_array = self._ag_to_mask()
    
    @property
    def name(self) -> str:
        return self._name
    
    @property
    def ui_item(self):
        return self.trajectory.object.mn_trajectory_selections[self.name]
    
    @property
    def selection_str(self) -> str:
        return self.ui_item.selection_str
    
    @selection_str.setter
    def selection_str(self, selection_str: str) -> None:
        self.ui_item.selection_str = selection_str
        try:
            self.set_atom_group(selection_str)
            self.set_selection()
            self.message = ""
        except Exception as e:
            self.message = str(e)
            print(e)
        
    @property
    def periodic(self) -> bool:
        return self.ui_item.periodic

    @periodic.setter
    def periodic(self, periodic: bool) -> None:
        if self.ui_item.periodic == periodic:
            return
        self.ui_item.periodic = periodic
    
    @property
    def updating(self) -> bool:
        return self.ui_item.updating
    
    @updating.setter
    def updating(self, updating: bool) -> None:
        if self.ui_item.updating == updating:
            return
        self.ui_item.updating = updating

    @property
    def message(self) -> str:
        return self.ui_item.message
    
    @message.setter
    def message(self, message: str) -> None:
        self.ui_item.message = message

    @property
    def immutable(self) -> bool:
        return self.ui_item.immutable
    
    @immutable.setter
    def immutable(self, immutable: bool) -> None:
        self.ui_item.immutable = immutable
    
    def set_atom_group(self, selection_str: str) -> None:
        if self._current_selection_str == selection_str:
            return
        try:
            self._ag = self.trajectory.universe.select_atoms(
                selection_str,
                updating=self.updating,
                periodic=self.periodic
            )
            self.mask_array = self._ag_to_mask()
            self._current_selection_str = selection_str
            self.message = ""
        except Exception as e:
            self.message = str(e)
            print(str(e) + f" in selection: `{self.name}` on object: `{self.trajectory.object.name}`" )
    
    def _ag_to_mask(self) -> npt.NDArray[np.bool_]:
        "Return a 1D boolean mask for the Universe atoms that are in the Selection's AtomGroup."
        return np.isin(self.trajectory.universe.atoms.ix, self._ag.ix).astype(bool)
    
    def set_selection(self) -> None:
        "Sets the selection in the trajectory"
        if not self.updating:
            return
        self.trajectory.set_boolean(self.to_mask(), name=self.name)

    def to_mask(self) -> npt.NDArray[np.bool_]:
        "Returns the selection as a 1D numpy boolean mask. If updating=True, recomputes selection."
        if self.updating:
            self.mask_array = self._ag_to_mask()
        return self.mask_array

    @classmethod
    def from_atomgroup(cls, trajectory, atomgroup: mda.AtomGroup, name: str = ""):
        "Create a Selection object from an AtomGroup"
        # set default value
        selection_str = f"sel_{atomgroup.n_atoms}_atoms"
        updating = False
        periodic = False

        # if class is an UpdatingAtomGroup
        if atomgroup.__class__.__name__ == "UpdatingAtomGroup":
            updating = True
            # assuming it's a single selection
            # MDA do support `u.select_atoms('index 0', 'around 5 index 0')
            selection_str = atomgroup._selection_strings[0]
            periodic = False
            try:
                if atomgroup._selections[0].periodic:
                    periodic = True
            except AttributeError as e:
                print(e)

        if name == "":
            name = selection_str
        selection = cls(trajectory, atomgroup.universe, "all", name, updating, periodic)

        selection.selection_str = selection_str
        selection.ag = atomgroup
        selection.mask_array = selection._ag_to_mask()
        return selection



import molecularnodes as mn
import bpy


def test_template():
    mn.template.install()
    bpy.ops.wm.read_homefile(app_template="Molecular Nodes")
    assert not bpy.data.objects.get("Cube")

    mn.template.uninstall()
    try:
        bpy.ops.wm.read_homefile(app_template="Molecular Nodes")
        assert False
    except Exception:
        assert True

    bpy.ops.wm.read_homefile(app_template="")



from os.path import join, dirname, realpath
import itertools
import pytest
import numpy as np
import biotite.structure.io.pdb as biotite_pdb
import biotite.structure.io.pdbx as biotite_cif
import molecularnodes.entities.molecule.pdb as pdb
import molecularnodes.entities.molecule.pdbx as pdbx

DATA_DIR = join(dirname(realpath(__file__)), "data")


@pytest.mark.parametrize(
    "pdb_id, format", itertools.product(["1f2n", "5zng"], ["pdb", "cif"])
)
def test_get_transformations(pdb_id, format):
    """
    Compare an assembly built from transformation information in
    MolecularNodes with assemblies built in Biotite.
    """
    path = join(DATA_DIR, f"{pdb_id}.{format}")
    if format == "pdb":
        pdb_file = biotite_pdb.PDBFile.read(path)
        atoms = biotite_pdb.get_structure(pdb_file, model=1)
        ref_assembly = biotite_pdb.get_assembly(pdb_file, model=1)
        test_parser = pdb.PDBAssemblyParser(pdb_file)
    elif format == "cif":
        cif_file = biotite_cif.CIFFile().read(path)
        atoms = biotite_cif.get_structure(
            # Make sure `label_asym_id` is used instead of `auth_asym_id`
            cif_file,
            model=1,
            use_author_fields=False,
        )
        ref_assembly = biotite_cif.get_assembly(cif_file, model=1)
        test_parser = pdbx.CIFAssemblyParser(cif_file)
    else:
        raise ValueError(f"Format '{format}' does not exist")

    assembly_id = test_parser.list_assemblies()[0]
    test_transformations = test_parser.get_transformations(assembly_id)

    if format == "pdb":
        check_transformations(test_transformations, atoms, ref_assembly)
    elif format == "cif":
        # note we need to use the new CIF parser but it returns a list of dicts which
        # needs to be converted to a list of 2-tuples
        test_transformations2 = [( transformation['chain_ids'], transformation['matrix']) for transformation in test_transformations]
        check_transformations(test_transformations2, atoms, ref_assembly)
    else:
        raise ValueError(f"Format '{format}' does not exist")


@pytest.mark.parametrize("assembly_id", [str(i + 1) for i in range(5)])
def test_get_transformations_cif(assembly_id):
    """
    Compare an assembly built from transformation information in
    MolecularNodes with assemblies built in Biotite.

    In this case all assemblies from a structure with more complex
    operation expressions are tested
    """
    cif_file = biotite_cif.CIFFile().read(join(DATA_DIR, "1f2n.cif"))
    atoms = biotite_cif.get_structure(
        # Make sure `label_asym_id` is used instead of `auth_asym_id`
        cif_file,
        model=1,
        use_author_fields=False,
    )
    ref_assembly = biotite_cif.get_assembly(cif_file, model=1, assembly_id=assembly_id)

    test_parser = pdbx.CIFAssemblyParser(cif_file)

    test_transformations = test_parser.get_transformations(assembly_id)

    test_transformations2 = [( transformation['chain_ids'], transformation['matrix']) for transformation in test_transformations]
    check_transformations(test_transformations2, atoms, ref_assembly)


def check_transformations(transformations, atoms, ref_assembly):
    """
    Check if the given transformations applied on the given atoms
    results in the given reference assembly.
    """
    test_assembly = None
    for chain_ids, matrix in transformations:
        matrix = np.array(matrix)
        translation = matrix[:3, 3]
        rotation = matrix[:3, :3]
        sub_assembly = atoms[np.isin(atoms.chain_id, chain_ids)].copy()
        sub_assembly.coord = np.dot(rotation, sub_assembly.coord.T).T
        sub_assembly.coord += translation
        if test_assembly is None:
            test_assembly = sub_assembly
        else:
            test_assembly += sub_assembly

    assert test_assembly.array_length() == ref_assembly.array_length()
    # The atom name is used as indicator of correct atom ordering here
    assert np.all(test_assembly.atom_name == ref_assembly.atom_name)
    assert np.allclose(test_assembly.coord, ref_assembly.coord, atol=1e-4)



import numpy as np
from MDAnalysis.topology.base import TopologyReaderBase
from MDAnalysis.core.topology import Topology
from MDAnalysis.core.topologyattrs import (
    Atomids,
    Bonds,
    ChainIDs,
    Resids,
    Resnames,
    Resnums,
)


class OXDNAParser(TopologyReaderBase):
    def parse(self, **kwargs):
        """Parse topology from an oxDNA topology file.

        Returns
        -------
        top : MDAnalysis.core.topology.Topology
            Topology object
        """
        top = self._parseatoms()

        return top

    @classmethod
    def _is_new_topology(cls, filename) -> bool:
        """Check if topology file is new format.

        Parameters
        ----------
        filename : str
            Path to topology file

        Returns
        -------
        bool
            True if topology file is new format
        """
        with open(filename) as f:
            return "5->3" in f.readline()

    @classmethod
    def _read_topo_new(cls, filename) -> Topology:
        """Read topology from new format oxDNA topology file.

        Parameters
        ----------
        filename : str
            Path to topology file

        Returns
        -------
        topo : MDAnalysis.core.topology.Topology
            Topology object
        """
        with open(filename) as f:
            lines = f.readlines()
        n_atoms, n_chains, direction = np.array(lines[0].split())
        n_atoms = int(n_atoms)
        atom_idx = np.arange(n_atoms)

        res_name_list = []
        chain_id_list = []
        for i, line in enumerate(lines[1:]):
            is_rna = "type=RNA" in line
            is_dna = not is_rna

            line_split = line.split()
            bases = line_split[0]

            base_list = []
            start = 0
            end = 0
            in_custom_base = False
            for i, letter in enumerate(bases):
                if letter == "(":
                    start = i + 1
                    in_custom_base = True
                    continue
                if letter == ")":
                    end = i
                    base_list.append(line[start:end])
                    in_custom_base = False
                    continue

                if in_custom_base:
                    continue

                base_list.append(letter)
            chain_id_list.append(np.repeat(i, len(base_list)))

            res_name_list.append(np.array(base_list))

        res_names = np.hstack(res_name_list)
        chain_ids = np.hstack(chain_id_list)

        bond_idx = np.zeros((n_atoms, 2), dtype=int)
        bond_idx[:, :] = -1

        for i in atom_idx:
            if i == 1:
                continue
            if chain_ids[i] == chain_ids[-1]:
                bond_idx[i, :] = np.array((i, i - 1), dtype=int)

        mask = np.logical_and(bond_idx[:, 0] != -1, bond_idx[:, 1] != -1)
        bond_idx = bond_idx[mask, :]

        attrs = [
            Atomids(atom_idx),
            Resids(atom_idx + 1),
            ChainIDs(chain_ids),
            Resnames(res_names),
        ]

        topo = Topology(
            n_atoms=n_atoms, n_res=n_atoms, attrs=attrs, atom_resindex=atom_idx
        )
        topo.add_TopologyAttr(Bonds(bond_idx))

        return topo

    @classmethod
    def _read_topo_old(cls, filename) -> Topology:
        """Read topology from old format oxDNA topology file.

        Parameters
        ----------
        filename : str
            Path to topology file

        Returns
        -------
        topo : MDAnalysis.core.topology.Topology
            Topology object
        """
        with open(filename) as f:
            first_line = f.readline()

        dimensions = np.array(first_line.split())
        n_atoms = int(dimensions[0])
        array = np.loadtxt(filename, skiprows=1, max_rows=n_atoms, dtype=str)

        # each topology item has two bond columns, which say what the base is bonded
        # from and what it is bonded to. -1 means it is not bonded
        # we need to turn that into an array of bond pairs
        bond_idx = np.zeros((n_atoms * 2, 2), int)
        row_numbers = np.arange(n_atoms, dtype=int)
        for i in range(2):
            rows = row_numbers + (i * n_atoms)
            bond_idx[rows, 0] = array[:, (i + 2)]
            bond_idx[rows, 1] = row_numbers

        # drop any that have -1 as they aren't bonded to anything
        mask = np.logical_and(bond_idx[:, 0] != -1, bond_idx[:, 1] != -1)
        bond_idx = bond_idx[mask, :]

        atom_idx = np.arange(n_atoms, dtype=int)
        attrs = [
            Atomids(atom_idx),
            Resids(atom_idx + 1),
            ChainIDs(array[:, 0].astype(int)),
            Resnames(array[:, 1].astype(str)),
        ]

        topo = Topology(
            n_atoms=n_atoms, n_res=n_atoms, attrs=attrs, atom_resindex=atom_idx
        )
        topo.add_TopologyAttr(Bonds(bond_idx))

        return topo

    def _parseatoms(self):
        """Parse atoms from topology file.

        Returns
        -------
        topo : MDAnalysis.core.topology.Topology
            Topology object
        """
        if self._is_new_topology(self.filename):
            return self._read_topo_new(self.filename)
        else:
            return self._read_topo_old(self.filename)



import colorsys
import random

import numpy as np
import numpy.typing as npt

import math


def clamp(value, min_value, max_value):
    return max(min_value, min(value, max_value))


class Lab:
    Kn = 18
    Xn = 0.950470
    Yn = 1
    Zn = 1.088830
    T0 = 0.137931034
    T1 = 0.206896552
    T2 = 0.12841855
    T3 = 0.008856452

    def __init__(self, l=0.1, a=0.0, b=0.0):
        self.lab = [l, a, b]
        self.l = l
        self.a = a
        self.b = b

    @staticmethod
    def zero():
        return Lab(0, 0, 0)

    @staticmethod
    def distance(a, b):
        x = b.l - a.l
        y = b.a - a.a
        z = b.b - a.b
        return math.sqrt(x * x + y * y + z * z)

    @staticmethod
    def darken(out, c, amount):
        out.l = c.l - Lab.Kn * amount
        out.a = c.a
        out.b = c.b
        return out

    @staticmethod
    def lighten(out, c, amount):
        return Lab.darken(out, c, -amount)

    @staticmethod
    def darken_color(c, amount):
        tmp_darken_lab = Lab.from_color(c)
        return Lab.to_color(Lab.darken(tmp_darken_lab, tmp_darken_lab, amount))

    @staticmethod
    def lighten_color(c, amount):
        return Lab.darken_color(c, -amount)

    @staticmethod
    def from_color(color):
        r, g, b, a = color * 255
        x, y, z = Lab.rgbToXyz(r, g, b)
        l = 116 * y - 16
        return Lab(l if l >= 0 else 0, 500 * (x - y), 200 * (y - z))

    @staticmethod
    def to_color(lab):
        y = (lab.l + 16) / 116
        x = y if math.isnan(lab.a) else y + lab.a / 500
        z = y if math.isnan(lab.b) else y - lab.b / 200

        y = Lab.Yn * Lab.lab_xyz(y)
        x = Lab.Xn * Lab.lab_xyz(x)
        z = Lab.Zn * Lab.lab_xyz(z)

        r = Lab.xyz_rgb(3.2404542 * x - 1.5371385 * y - 0.4985314 * z)
        g = Lab.xyz_rgb(-0.9692660 * x + 1.8760108 * y + 0.0415560 * z)
        b = Lab.xyz_rgb(0.0556434 * x - 0.2040259 * y + 1.0572252 * z)

        return [
            round(clamp(r, 0, 255)) / 255.0,
            round(clamp(g, 0, 255)) / 255.0,
            round(clamp(b, 0, 255)) / 255.0,
            1.0,
        ]

    @staticmethod
    def xyz_rgb(c):
        return 255 * (
            12.92 * c if c <= 0.00304 else 1.055 * math.pow(c, 1 / 2.4) - 0.055
        )

    @staticmethod
    def lab_xyz(t):
        return t * t * t if t > Lab.T1 else Lab.T2 * (t - Lab.T0)

    @staticmethod
    def rgb_xyz(c):
        c /= 255
        return c / 12.92 if c <= 0.04045 else math.pow((c + 0.055) / 1.055, 2.4)

    @staticmethod
    def xyz_lab(t):
        return math.pow(t, 1 / 3) if t > Lab.T3 else t / Lab.T2 + Lab.T0

    @staticmethod
    def rgbToXyz(r, g, b):
        r = Lab.rgb_xyz(r)
        g = Lab.rgb_xyz(g)
        b = Lab.rgb_xyz(b)
        x = Lab.xyz_lab((0.4124564 * r + 0.3575761 * g + 0.1804375 * b) / Lab.Xn)
        y = Lab.xyz_lab((0.2126729 * r + 0.7151522 * g + 0.0721750 * b) / Lab.Yn)
        z = Lab.xyz_lab((0.0193339 * r + 0.1191920 * g + 0.9503041 * b) / Lab.Zn)
        return [x, y, z]


def random_rgb(seed=None):
    """Random Pastel RGB values"""
    if seed:
        random.seed(seed)
    r, g, b = colorsys.hls_to_rgb(random.random(), 0.6, 0.6)
    return np.array((r, g, b, 1))


def plddt(b_factor: np.ndarray) -> npt.NDArray[np.float32]:
    colors = np.zeros((len(b_factor), 4), float)

    for i, value in enumerate(b_factor):
        if value > 90:
            color = np.array([0.000000, 0.086496, 0.672395, 1.000000])
        elif value > 70:
            color = np.array([0.130157, 0.597176, 0.896205, 1.000000])
        elif value > 50:
            color = np.array([1.000169, 0.708345, 0.006512, 1.000000])
        else:
            color = np.array([1.000169, 0.205070, 0.059507, 1.000000])
        colors[i, :] = color

    return colors


def color_from_atomic_number(atomic_number: int):
    r, g, b = list(iupac_colors_rgb.values())[int(atomic_number - 1)]
    return np.array((r, g, b, 1))


def color_from_element(element: str):
    r, g, b = [x / 255 for x in iupac_colors_rgb[element]]
    return np.array((r, g, b, 1))


def colors_from_elements(atomic_numbers):
    colors = np.array(list(map(color_from_atomic_number, atomic_numbers)))
    return colors


def equidistant_colors(some_list):
    u = np.unique(some_list)
    num_colors = len(u)

    hues = [i / (num_colors + 1) for i in range(num_colors)]

    # Convert HSL to RGB
    colors = [colorsys.hls_to_rgb(hue, 0.6, 0.6) for hue in hues]

    # Convert RGB to 8-bit integer values
    colors = [(int(r * 255), int(g * 255), int(b * 255), 1) for (r, g, b) in colors]

    return dict(zip(u, colors))


def color_chains_equidistant(chain_ids):
    color_dict = equidistant_colors(chain_ids)
    chain_colors = np.array([color_dict[x] for x in chain_ids])
    return chain_colors / 255


def color_chains(atomic_numbers, chain_ids):
    mask = atomic_numbers == 6
    colors = colors_from_elements(atomic_numbers)
    chain_color_dict = equidistant_colors(chain_ids)
    chain_colors = np.array(list(map(lambda x: chain_color_dict.get(x), chain_ids)))

    colors[mask] = chain_colors[mask]

    return colors / 255


iupac_colors_rgb = {
    "H": (255, 255, 255),  # Hydrogen
    "He": (217, 255, 255),  # Helium
    "Li": (204, 128, 255),  # Lithium
    "Be": (194, 255, 0),  # Beryllium
    "B": (255, 181, 181),  # Boron
    "C": (144, 144, 144),  # Carbon
    "N": (48, 80, 248),  # Nitrogen
    "O": (255, 13, 13),  # Oxygen
    "F": (144, 224, 80),  # Fluorine
    "Ne": (179, 227, 245),  # Neon
    "Na": (171, 92, 242),  # Sodium
    "Mg": (138, 255, 0),  # Magnesium
    "Al": (191, 166, 166),  # Aluminum
    "Si": (240, 200, 160),  # Silicon
    "P": (255, 128, 0),  # Phosphorus
    "S": (255, 255, 48),  # Sulfur
    "Cl": (31, 240, 31),  # Chlorine
    "K": (143, 64, 212),  # Potassium
    "Ar": (128, 209, 227),  # Argon
    "Ca": (61, 255, 0),  # Calcium
    "Sc": (230, 230, 230),  # Scandium
    "Ti": (191, 194, 199),  # Titanium
    "V": (166, 166, 171),  # Vanadium
    "Cr": (138, 153, 199),  # Chromium
    "Mn": (156, 122, 199),  # Manganese
    "Fe": (224, 102, 51),  # Iron
    "Ni": (199, 138, 138),  # Nickel
    "Co": (255, 217, 143),  # Cobalt
    "Cu": (200, 128, 51),  # Copper
    "Zn": (125, 128, 176),  # Zinc
    "Ga": (194, 143, 143),  # Gallium
    "Ge": (102, 143, 143),  # Germanium
    "As": (189, 128, 227),  # Arsenic
    "Se": (255, 161, 0),  # Selenium
    "Br": (166, 41, 41),  # Bromine
    "Kr": (92, 184, 209),  # Krypton
    "Rb": (112, 46, 176),  # Rubidium
    "Sr": (0, 255, 0),  # Strontium
    "Y": (148, 255, 255),  # Yttrium
    "Zr": (148, 224, 224),  # Zirconium
    "Nb": (115, 194, 201),  # Niobium
    "Mo": (84, 181, 181),  # Molybdenum
    "Tc": (59, 158, 158),  # Technetium
    "Ru": (36, 125, 125),  # Ruthenium
    "Rh": (10, 125, 140),  # Rhodium
    "Pd": (0, 105, 133),  # Palladium
    "Ag": (192, 192, 192),  # Silver
    "Cd": (255, 217, 143),  # Cadmium
    "In": (166, 117, 115),  # Indium
    "Sn": (102, 128, 128),  # Tin
    "Sb": (158, 99, 181),  # Antimony
    "Te": (212, 122, 0),  # Tellurium
    "I": (148, 0, 148),  # Iodine
    "Xe": (66, 158, 176),  # Xenon
    "Cs": (87, 23, 143),  # Cesium
    "Ba": (0, 201, 0),  # Barium
    "La": (112, 212, 255),  # Lanthanum
    "Ce": (255, 255, 199),  # Cerium
    "Pr": (217, 255, 199),  # Praseodymium
    "Nd": (199, 255, 199),  # Neodymium
    "Pm": (163, 255, 199),  # Promethium
    "Sm": (143, 255, 199),  # Samarium
    "Eu": (97, 255, 199),  # Europium
    "Gd": (69, 255, 199),  # Gadolinium
    "Tb": (48, 255, 199),  # Terbium
    "Dy": (31, 255, 199),  # Dysprosium
    "Ho": (0, 255, 156),  # Holmium
    "Er": (0, 230, 117),  # Erbium
    "Tm": (0, 212, 82),  # Thulium
    "Yb": (0, 191, 56),  # Ytterbium
    "Lu": (0, 171, 36),  # Lutetium
    "Hf": (77, 194, 255),  # Hafnium
    "Ta": (77, 166, 255),  # Tantalum
    "W": (33, 148, 214),  # Tungsten
    "Re": (38, 125, 171),  # Rhenium
    "Os": (38, 102, 150),  # Osmium
    "Ir": (23, 84, 135),  # Iridium
    "Pt": (208, 208, 224),  # Platinum
    "Au": (255, 209, 35),  # Gold
    "Hg": (184, 184, 208),  # Mercury
    "Tl": (166, 84, 77),  # Thallium
    "Pb": (87, 89, 97),  # Lead
    "Bi": (158, 79, 181),  # Bismuth
    "Th": (255, 161, 0),  # Thorium
    "Pa": (255, 161, 0),  # Protactinium
    "U": (255, 161, 0),  # Uranium
    "Np": (255, 161, 0),  # Neptunium
    "Pu": (255, 161, 0),  # Plutonium
    "Am": (255, 161, 0),  # Americium
    "Cm": (255, 161, 0),  # Curium
    "Bk": (255, 161, 0),  # Berkelium
    "Cf": (255, 161, 0),  # Californium
    "Es": (255, 161, 0),  # Einsteinium
    "Fm": (255, 161, 0),  # Fermium
    "Md": (255, 161, 0),  # Mendelevium
    "No": (255, 161, 0),  # Nobelium
    "Lr": (255, 161, 0),  # Lawrencium
    "Rf": (204, 0, 89),  # Rutherfordium
    "Db": (209, 0, 79),  # Dubnium
    "Sg": (217, 0, 69),  # Seaborgium
    "Bh": (224, 0, 56),  # Bohrium
    "Hs": (230, 0, 46),  # Hassium
    "Mt": (235, 0, 38),  # Meitnerium
    "Ds": (240, 0, 33),  # Darmstadtium
    "Rg": (241, 0, 30),  # Roentgenium
    "Cn": (242, 0, 26),  # Copernicium
    "Nh": (242, 0, 26),  # Nihonium
    "Fl": (242, 0, 26),  # Flerovium
    "Mc": (242, 0, 26),  # Moscovium
    "Lv": (242, 0, 26),  # Livermorium
    "Ts": (242, 0, 26),  # Tennessine
    "Og": (242, 0, 26),  # Oganesson
}



from bpy.types import Material
from databpy.material import append_from_blend
from ..utils import MN_DATA_FILE

MATERIAL_NAMES = [
    "MN Default",
    "MN Flat Outline",
    "MN Squishy",
    "MN Transparent Outline",
    "MN Ambient Occlusion",
]


def append(name: str) -> Material:
    "Append a material from the MN_DATA_FILE."
    return append_from_blend(name, MN_DATA_FILE)


def add_all_materials() -> None:
    "Append all pre-defined materials from the MN_DATA_FILE."
    for name in MATERIAL_NAMES:
        append(name)


def default() -> Material:
    "Return the default material."
    return append("MN Default")


# class to interact with a bpy.types.Material node tree and change some of the default
# values of the nodes inside of it
class MaterialTreeInterface:
    def __init__(self, tree: Material):
        self.tree = tree
        self.nodes = tree.node_tree.nodes
        self.links = tree.node_tree.links


class MaterialAmbientOcclusion(MaterialTreeInterface):
    def __init__(self, tree: Material):
        super().__init__(tree)

    @property
    def ao_power(self) -> float:
        return self.nodes["Math"].inputs[1].default_value

    @ao_power.setter
    def ao_power(self, value: float):
        self.nodes["Math"].inputs[1].default_value = value

    @property
    def ao_distance(self) -> float:
        return self.nodes["Ambient Occlusion"].inputs["Distance"].default_value

    @ao_distance.setter
    def ao_distance(self, value: float) -> None:
        self.nodes["Ambient Occlusion"].inputs["Distance"].default_value = value


class MaterialSquishy(MaterialTreeInterface):
    def __init__(self, tree: Material):
        super().__init__(tree)

    @property
    def subsurf_scale(self) -> float:
        return self.nodes["Principled BSDF"].inputs["Subsurface Scale"].default_value

    @subsurf_scale.setter
    def subsurf_scale(self, value: float):
        self.nodes["Principled BSDF"].inputs["Subsurface Scale"].default_value = value



from .base import Density

import mrcfile
from ...blender import coll, nodes
import databpy
import bpy
import numpy as np
import os


class MRC(Density):
    """
    A class for parsing EM density files in the format `.map` or `.map.gz`.

    It utilises `mrcfile` for file parsing, which is then converted into `pyopevdb` grids,
    that can be written as `.vdb` files and the imported into Blender as volumetric objects.
    """

    def __init__(self, file_path, center=False, invert=False, overwrite=False):
        super().__init__(file_path=file_path)
        self.grid = self.map_to_grid(self.file_path, center=center)
        self.file_vdb = self.map_to_vdb(
            self.file_path, center=center, invert=invert, overwrite=overwrite
        )

    def create_object(
        self, name="NewDensity", style="density_surface", setup_nodes=True
    ) -> bpy.types.Object:
        """
        Loads an MRC file into Blender as a volumetric object.

        Parameters
        ----------
        file : str
            Path to the MRC file.
        name : str, optional
            If not None, renames the object with the new name.

        Returns
        -------
        bpy.types.Object
            The loaded volumetric object.
        """
        # import and ensure object is at world origin to get corect alignment with
        # structures
        self.object = databpy.import_vdb(self.file_vdb, collection=coll.mn())
        self.object.location = (0, 0, 0)
        self.object.mn.entity_type = self._entity_type.value

        if name and name != "":
            self.name = name

        self.create_starting_node_tree(style=style)

        return self.object

    def create_starting_node_tree(self, style="density_surface"):
        """
        Creates a starting node tree for the density object.

        Parameters
        ----------
        style : str, optional
            The style of the density object, defaulting to 'density_surface'.
        """
        nodes.create_starting_nodes_density(
            object=self.object, style=style, threshold=self.threshold
        )

    def map_to_vdb(
        self,
        file: str,
        invert: bool = False,
        world_scale=0.01,
        center: bool = False,
        overwrite=False,
    ) -> (str, float):
        """
        Converts an MRC file to a .vdb file using pyopenvdb.

        Parameters
        ----------
        file : str
            The path to the input MRC file.
        invert : bool, optional
            Whether to invert the data from the grid, defaulting to False. Some file types
            such as EM tomograms have inverted values, where a high value == low density.
        world_scale : float, optional
            The scaling factor to apply to the voxel size of the input file. Defaults to 0.01.
        center : bool, optional
            Whether to center the volume on the origin. Defaults to False.
        overwrite : bool, optional
            If True, the .vdb file will be overwritten if it already exists. Defaults to False.

        Returns
        -------
        str
            The path to the converted .vdb file.
        """
        import pyopenvdb as vdb

        file_path = self.path_to_vdb(file, center=center, invert=invert)

        # If the map has already been converted to a .vdb and overwrite is False, return that instead
        if os.path.exists(file_path) and not overwrite:
            # Also check that the file has the same invert and center settings
            grid = vdb.readAllGridMetadata(file_path)[0]
            if (
                "MN_invert" in grid
                and grid["MN_invert"] == invert
                and "MN_center" in grid
                and grid["MN_center"] == center
            ):
                self.threshold = grid["MN_initial_threshold"]
                return file_path

        print("Reading new file")
        # Read in the MRC file and convert it to a pyopenvdb grid
        grid = self.map_to_grid(file=file, invert=invert, center=center)

        grid.transform.scale(np.array((1, 1, 1)) * world_scale * grid["MN_voxel_size"])

        if center:
            offset = -np.array(grid["MN_box_size"]) * 0.5
            offset *= grid["MN_voxel_size"] * world_scale
            print("transforming")
            grid.transform.translate(offset)

        if os.path.exists(file_path):
            os.remove(file_path)

        # Write the grid to a .vdb file
        print("writing new file")
        vdb.write(file_path, grids=[grid])
        self.threshold = grid["MN_initial_threshold"]
        del grid

        # Return the path to the output file
        return file_path

    def map_to_grid(self, file: str, invert: bool = False, center: bool = False):
        """
        Reads an MRC file and converts it into a pyopenvdb FloatGrid object.

        This function reads a file in MRC format, and converts it into a pyopenvdb FloatGrid object,
        which can be used to represent volumetric data in Blender.

        Parameters
        ----------
        file : str
            The path to the MRC file.
        invert : bool, optional
            Whether to invert the data from the grid, defaulting to False. Some file types
            such as EM tomograms have inverted values, where a high value == low density.

        Returns
        -------
        pyopenvdb.FloatGrid
            A pyopenvdb FloatGrid object containing the density data.
        """

        import pyopenvdb as vdb

        volume = mrcfile.read(file)

        dataType = volume.dtype

        # enables different grid types

        if dataType == "float32" or dataType == "float64":
            grid = vdb.FloatGrid()
        elif dataType == "int8" or dataType == "int16" or dataType == "int32":
            volume = volume.astype("int32")
            grid = vdb.Int32Grid()
        elif dataType == "int64":
            grid = vdb.Int64Grid()
        else:
            grid = vdb.FloatGrid()

        if invert:
            volume = np.max(volume) - volume

        initial_threshold = np.quantile(volume, 0.995)

        # The np.copy is needed to force numpy to actually rewrite the data in memory
        # since openvdb seems to read is straight from memory without checking the striding
        # The np.transpose is needed to convert the data from zyx to xyz
        volume = np.copy(np.transpose(volume, (2, 1, 0)), order="C")
        try:
            grid.copyFromArray(volume.astype(float))
        except Exception as e:
            print(
                f"Grid data type '{volume.dtype}' is an unsupported type.\nError: {e}"
            )

        grid.gridClass = vdb.GridClass.FOG_VOLUME
        grid.name = "density"

        # Set some metadata for the vdb file, so we can check if it's already been converted
        # correctly
        grid["MN_invert"] = invert
        grid["MN_initial_threshold"] = initial_threshold
        grid["MN_center"] = center
        with mrcfile.open(file) as mrc:
            grid["MN_voxel_size"] = float(mrc.voxel_size.x)
            grid["MN_box_size"] = (
                int(mrc.header.nx),
                int(mrc.header.ny),
                int(mrc.header.nz),
            )

        return grid



import itertools
import math
import os
from typing import List, Optional

import bpy
import databpy
import numpy as np
from databpy.nodes import (
    NodeGroupCreationError,
    append_from_blend,
    swap_tree,
)

from .. import color, utils
from ..utils import MN_DATA_FILE
from . import material, mesh

NODE_WIDTH = 180


socket_types = {
    "BOOLEAN": "NodeSocketBool",
    "GEOMETRY": "NodeSocketGeometry",
    "INT": "NodeSocketInt",
    "MATERIAL": "NodeSocketMaterial",
    "VECTOR": "NodeSocketVector",
    "STRING": "NodeSocketString",
    "VALUE": "NodeSocketFloat",
    "COLLECTION": "NodeSocketCollection",
    "TEXTURE": "NodeSocketTexture",
    "COLOR": "NodeSocketColor",
    "RGBA": "NodeSocketColor",
    "IMAGE": "NodeSocketImage",
}

# current implemented representations
styles_mapping = {
    "preset_1": "Style Preset 1",
    "preset_2": "Style Preset 2",
    "preset_3": "Style Preset 3",
    "preset_4": "Style Preset 4",
    "atoms": "Style Spheres",
    "spheres": "Style Spheres",
    "vdw": "Style Spheres",
    "sphere": "Style Spheres",
    "cartoon": "Style Cartoon",
    "sticks": "Style Sticks",
    "ribbon": "Style Ribbon",
    "surface": "Style Surface",
    "ball_and_stick": "Style Ball and Stick",
    "ball+stick": "Style Ball and Stick",
    "oxdna": "MN_oxdna_style_ribbon",
    "density_surface": "Style Density Surface",
    "density_wire": "Style Density Wire",
}


def inputs(node):
    items = {}
    for item in node.interface.items_tree:
        if item.item_type == "SOCKET":
            if item.in_out == "INPUT":
                items[item.name] = item
    return items


def outputs(node):
    items = {}
    for item in node.interface.items_tree:
        if item.item_type == "SOCKET":
            if item.in_out == "OUTPUT":
                items[item.name] = item
    return items


def get_output_type(node, type="INT"):
    for output in node.outputs:
        if output.type == type:
            return output


def set_selection(group, node, selection):
    pos = node.location
    pos = [pos[0] - 200, pos[1] - 200]
    selection.location = pos
    group.links.new(selection.outputs[0], node.inputs["Selection"])

    return selection


def create_debug_group(name="MolecularNodesDebugGroup"):
    group = new_tree(name=name, fallback=False)
    info = group.nodes.new("GeometryNodeObjectInfo")
    group.links.new(info.outputs["Geometry"], group.nodes["Group Output"].inputs[0])
    return group


def add_selection(group, sel_name, input_list, field="chain_id"):
    style = style_node(group)
    sel_node = add_custom(
        group,
        custom_iswitch(
            name="selection", iter_list=input_list, field=field, dtype="BOOLEAN"
        ).name,
    )

    set_selection(group, style, sel_node)
    return sel_node


def get_output(group):
    return group.nodes[
        bpy.app.translations.pgettext_data(
            "Group Output",
        )
    ]


def get_input(group):
    return group.nodes[
        bpy.app.translations.pgettext_data(
            "Group Input",
        )
    ]


def get_mod(object, name="MolecularNodes"):
    node_mod = object.modifiers.get(name)
    if not node_mod:
        node_mod = object.modifiers.new(name, "NODES")
    object.modifiers.active = node_mod
    return node_mod


def get_nodes_last_output(group):
    output = get_output(group)
    last = output.inputs[0].links[0].from_node
    return last, output


def previous_node(node):
    "Get the node which is the first connection to the first input of this node"
    prev = node.inputs[0].links[0].from_node
    return prev


def style_node(group):
    prev = previous_node(get_output(group))
    is_style_node = "Style" in prev.name
    while not is_style_node:
        prev = previous_node(prev)
        is_style_node = "Style" in prev.name
    return prev


def get_style_node(object):
    "Walk back through the primary node connections until you find the first style node"
    group = object.modifiers["MolecularNodes"].node_group
    return style_node(group)


def star_node(group):
    prev = previous_node(get_output(group))
    is_star_node = "Starfile Instances" in prev.name
    while not is_star_node:
        prev = previous_node(prev)
        is_star_node = "Starfile Instances" in prev.name
    return prev


def get_star_node(object):
    "Walk back through the primary node connections until you find the first style node"
    group = object.modifiers["MolecularNodes"].node_group
    return star_node(group)


def get_color_node(object):
    "Walk back through the primary node connections until you find the first style node"
    group = object.modifiers["MolecularNodes"].node_group
    for node in group.nodes:
        if node.name == "Color Attribute Random":
            return node


def insert_last_node(group, node, link_input=True):
    last, output = get_nodes_last_output(group)
    link = group.links.new
    location = output.location
    output.location = [location[0] + 300, location[1]]
    node.location = [location[0] - 300, location[1]]
    if link_input:
        link(last.outputs[0], node.inputs[0])
    link(node.outputs[0], output.inputs[0])


def realize_instances(obj):
    group = obj.modifiers["MolecularNodes"].node_group
    realize = group.nodes.new("GeometryNodeRealizeInstances")
    insert_last_node(group, realize)


def swap(node: bpy.types.GeometryNode, tree: str | bpy.types.GeometryNodeTree) -> None:
    "Swap out the node's node_tree, while maintaining the possible old connections"

    if isinstance(tree, str):
        try:
            tree = bpy.data.node_groups[tree]
        except KeyError:
            tree = append(tree)

    swap_tree(node=node, tree=tree)


def append(name: str, link: bool = False) -> bpy.types.GeometryNodeTree:
    "Append a GN node from the MN data file"
    GN_TREES_PATH = os.path.join(MN_DATA_FILE, "NodeTree")
    return append_from_blend(name, filepath=GN_TREES_PATH, link=link)


def MN_micrograph_material():
    """
    Append MN_micrograph_material to the .blend file it it doesn't already exist,
    and return that material.
    """

    mat_name = "MN_micrograph_material"

    return bpy.data.materials[mat_name]


def new_tree(
    name: str = "Geometry Nodes",
    geometry: bool = True,
    input_name: str = "Geometry",
    output_name: str = "Geometry",
    fallback: bool = True,
) -> bpy.types.GeometryNodeTree:
    group = bpy.data.node_groups.get(name)
    # if the group already exists, return it and don't create a new one
    if group and fallback:
        return group

    # create a new group for this particular name and do some initial setup
    group = bpy.data.node_groups.new(name, "GeometryNodeTree")
    input_node = group.nodes.new("NodeGroupInput")
    output_node = group.nodes.new("NodeGroupOutput")
    input_node.location.x = -200 - input_node.width
    output_node.location.x = 200
    if geometry:
        group.interface.new_socket(
            input_name, in_out="INPUT", socket_type="NodeSocketGeometry"
        )
        group.interface.new_socket(
            output_name, in_out="OUTPUT", socket_type="NodeSocketGeometry"
        )
        group.links.new(output_node.inputs[0], input_node.outputs[0])
    return group


def assign_material(node, new_material="default") -> None:
    material.add_all_materials()
    material_socket = node.inputs.get("Material")
    if material_socket is None:
        return None

    if isinstance(new_material, bpy.types.Material):
        material_socket.default_value = new_material
    elif new_material == "default":
        material_socket.default_value = material.default()
    else:
        try:
            material_socket.default_value = material.append(new_material)
        except Exception as e:
            print(f"Unable to use material {new_material}, error: {e}")


def add_custom(
    group,
    name,
    location=[0, 0],
    width=NODE_WIDTH,
    material="default",
    show_options=False,
    link=False,
):
    node = group.nodes.new("GeometryNodeGroup")
    node.node_tree = append(name, link=link)

    # if there is an input socket called 'Material', assign it to the base MN material
    # if another material is supplied, use that instead.
    assign_material(node, new_material=material)

    # move and format the node for arranging
    node.location = location
    node.width = width
    node.show_options = show_options
    node.name = name

    return node


def change_style_node(obj: bpy.types.Object, style: str):
    swap(get_style_node(obj), append(styles_mapping[style]))


def create_starting_nodes_starfile(object):
    # ensure there is a geometry nodes modifier called 'MolecularNodes' that is created and applied to the object
    node_mod = get_mod(object)

    node_name = f"MN_starfile_{object.name}"

    # create a new GN node group, specific to this particular molecule
    group = new_tree(node_name)
    node_mod.node_group = group
    link = group.links.new

    # move the input and output nodes for the group
    node_input = get_input(group)
    node_output = get_output(group)
    node_input.location = [0, 0]
    node_output.location = [700, 0]
    node_star_instances = add_custom(group, "Starfile Instances", [450, 0])
    link(node_star_instances.outputs[0], node_output.inputs[0])
    link(node_input.outputs[0], node_star_instances.inputs[0])


def create_starting_nodes_density(object, threshold=0.8, style="density_surface"):
    # ensure there is a geometry nodes modifier called 'MolecularNodes' that is created and applied to the object
    mod = get_mod(object)
    node_name = f"MN_density_{object.name}"

    try:
        tree = bpy.data.node_groups[node_name]
        mod.node_group = tree
        return
    except KeyError:
        pass

    # create a new GN node group, specific to this particular molecule
    group = new_tree(node_name, fallback=False)
    link = group.links.new
    mod.node_group = group

    # move the input and output nodes for the group
    node_input = get_input(group)
    node_input.location = [0, 0]
    node_output = get_output(group)
    node_output.location = [800, 0]

    node_density = add_custom(group, styles_mapping[style], [400, 0])
    node_density.inputs["Threshold"].default_value = threshold

    link(node_input.outputs[0], node_density.inputs[0])
    link(node_density.outputs[0], node_output.inputs[0])


def create_starting_node_tree(
    object: bpy.types.Object,
    coll_frames: bpy.types.Collection | None = None,
    style: str = "spheres",
    name: str | None = None,
    color: str = "common",
    material: str = "MN Default",
    is_modifier: bool = True,
):
    """
    Create a starting node tree for the inputted object.

    Parameters
    ----------
    object : bpy.types.Object
        Object to create the node tree for.
    coll_frames : bpy.data.collections, optional
        If None, no animation will be created.
        The default is None.
    style : str, optional
        Starting style for the node tree. The default is "spheres".
        Available options are stored as the keys of styles_mapping.
    name : str, optional
        Name of the node tree. If None, a default name will be generated.
        The default is None.
    color : str, optional
        None doesn't add ay set_color nodes, 'common' adds the color by common elements
        and 'plddt' adds color by pLDDT score.
    """
    # ensure there is a geometry nodes modifier called 'MolecularNodes' that is created and applied to the object
    mod = get_mod(object)

    if not name:
        name = f"MN_{object.name}"

    # check if the node tree already exists and use that instead
    try:
        tree = bpy.data.node_groups[name]
        mod.node_group = tree
        return
    except KeyError:
        pass

    tree = new_tree(name, input_name="Atoms")
    tree.is_modifier = is_modifier
    link = tree.links.new
    mod.node_group = tree

    # move the input and output nodes for the group
    node_input = get_input(tree)
    node_output = get_output(tree)
    node_input.location = [0, 0]
    node_output.location = [700, 0]

    node_style = add_custom(tree, styles_mapping[style], [450, 0], material=material)
    link(node_style.outputs[0], node_output.inputs[0])
    link(node_input.outputs[0], node_style.inputs[0])

    # if requested, setup the nodes for generating colors in the node tree
    if color is not None:
        if color == "common":
            node_color_set = add_custom(tree, "Set Color", [200, 0])
            node_color_common = add_custom(tree, "Color Common", [-50, -150])
            node_random_color = add_custom(tree, "Color Attribute Random", [-300, -150])

            link(node_input.outputs[0], node_color_set.inputs[0])
            link(node_random_color.outputs["Color"], node_color_common.inputs["Carbon"])
            link(node_color_common.outputs[0], node_color_set.inputs["Color"])
            link(node_color_set.outputs[0], node_style.inputs[0])
            to_animate = node_color_set
        elif color.lower() == "plddt":
            node_color_set = add_custom(tree, "Set Color", [200, 0])
            node_color_plddt = add_custom(tree, "Color pLDDT", [-50, -150])

            link(node_input.outputs[0], node_color_set.inputs["Atoms"])
            link(node_color_plddt.outputs[0], node_color_set.inputs["Color"])
            link(node_color_set.outputs["Atoms"], node_style.inputs["Atoms"])
        else:
            to_animate = node_input
    else:
        to_animate = node_input

    # if multiple frames, set up the required nodes for an animation
    if coll_frames:
        node_output.location = [1100, 0]
        node_style.location = [800, 0]

        node_animate_frames = add_custom(tree, "Animate Frames", [500, 0])
        node_animate = add_custom(tree, "Animate Value", [500, -300])

        node_animate_frames.inputs["Frames"].default_value = coll_frames
        node_animate.inputs["Value Max"].default_value = len(coll_frames.objects) - 1

        link(to_animate.outputs[0], node_animate_frames.inputs[0])
        link(node_animate_frames.outputs[0], node_style.inputs[0])
        link(node_animate.outputs[0], node_animate_frames.inputs["Frame"])


def combine_join_geometry(group, node_list, output="Geometry", join_offset=300):
    link = group.links.new
    max_x = max([node.location[0] for node in node_list])
    node_to_instances = group.nodes.new("GeometryNodeJoinGeometry")
    node_to_instances.location = [int(max_x + join_offset), 0]

    for node in reversed(node_list):
        link(node.outputs[output], node_to_instances.inputs["Geometry"])
    return node_to_instances


def split_geometry_to_instances(name, iter_list=("A", "B", "C"), attribute="chain_id"):
    """Create a Node to Split Geometry by an Attribute into Instances

    Splits the inputted geometry into instances, based on an attribute field. By
    default this field is the `chain_id` but this can be selected for any field.
    Will loop over each item of the list, so a list of arbitrary items that will
    define how many times to create the required nodes.

    """
    group = new_tree(name)
    node_input = get_input(group)
    node_output = get_output(group)

    named_att = group.nodes.new("GeometryNodeInputNamedAttribute")
    named_att.location = [-200, -200]
    named_att.data_type = "INT"
    named_att.inputs[0].default_value = attribute

    link = group.links.new
    list_sep = []

    for i, chain in enumerate(iter_list):
        pos = [i % 10, math.floor(i / 10)]

        node_split = add_custom(group, ".MN_utils_split_instance")
        node_split.location = [int(250 * pos[0]), int(-300 * pos[1])]
        node_split.inputs["Group ID"].default_value = i
        link(named_att.outputs["Attribute"], node_split.inputs["Field"])
        link(node_input.outputs[0], node_split.inputs["Geometry"])
        list_sep.append(node_split)

    node_instance = combine_join_geometry(group, list_sep, "Instance")
    node_output.location = [int(10 * 250 + 400), 0]
    link(node_instance.outputs[0], node_output.inputs[0])
    return group


def assembly_initialise(obj: bpy.types.Object):
    """
    Setup the required data object and nodes for building an assembly.
    """

    data_obj_name = f".data_assembly_{obj.name}"

    # check if a data object exists and create a new one if not
    data_object = bpy.data.objects.get(data_obj_name)
    if not data_object:
        transforms = utils.array_quaternions_from_dict(obj.mn.biological_assemblies)
        data_object = mesh.create_data_object(array=transforms, name=data_obj_name)

    tree_assembly = create_assembly_node_tree(name=obj.name, data_object=data_object)
    return tree_assembly


def assembly_insert(mol: bpy.types.Object):
    """
    Given a molecule, setup the required assembly node and insert it into the node tree.
    """

    tree_assembly = assembly_initialise(mol)
    group = get_mod(mol).node_group
    node = add_custom(group, tree_assembly.name)
    insert_last_node(get_mod(mol).node_group, node)


def create_assembly_node_tree(
    name: str, data_object: bpy.types.Object
) -> bpy.types.NodeTree:
    node_group_name = f"Assembly {name}"
    existing_node_tree = bpy.data.node_groups.get(node_group_name)
    if existing_node_tree:
        return existing_node_tree

    tree: bpy.types.NodeTree = new_tree(name=node_group_name)
    link = tree.links.new

    node_split = add_custom(tree, "Split to Centred Instances", [-150, 0])

    node_att: bpy.types.GeometryNodeInputNamedAttribute = tree.nodes.new(
        "GeometryNodeInputNamedAttribute"
    )  # type: ignore
    node_att.data_type = "INT"
    node_att.inputs[0].default_value = "chain_id"
    node_att.location = [-150, -200]
    link(node_att.outputs["Attribute"], node_split.inputs["Group ID"])

    node_group_assembly_instance = append(".MN_assembly_instance_chains")
    node_assembly = add_custom(tree, node_group_assembly_instance.name, [150, 0])
    node_assembly.inputs["data_object"].default_value = data_object

    out_sockets = outputs(tree)
    out_sockets[list(out_sockets)[0]].name = "Instances"

    socket_info = (
        {
            "name": "Rotation",
            "type": "NodeSocketFloat",
            "min": 0,
            "max": 1,
            "default": 1,
        },
        {
            "name": "Translation",
            "type": "NodeSocketFloat",
            "min": 0,
            "max": 1,
            "default": 1,
        },
        {
            "name": "assembly_id",
            "type": "NodeSocketInt",
            "min": 1,
            "max": max(databpy.named_attribute(data_object, "assembly_id")),
            "default": 1,
        },
    )

    for info in socket_info:
        socket = tree.interface.items_tree.get(info["name"])
        if not socket:
            socket: bpy.types.NodeTreeInterfaceSocket = tree.interface.new_socket(
                info["name"], in_out="INPUT", socket_type=info["type"]
            )
        socket.default_value = info["default"]
        socket.min_value = info["min"]
        socket.max_value = info["max"]

        link(get_input(tree).outputs[info["name"]], node_assembly.inputs[info["name"]])

    get_output(tree).location = [400, 0]
    link(get_input(tree).outputs[0], node_split.inputs[0])
    link(node_split.outputs[0], node_assembly.inputs[0])
    link(node_assembly.outputs[0], get_output(tree).inputs[0])
    if hasattr(tree, "color_tag"):
        tree.color_tag = "GEOMETRY"
    return tree


def add_inverse_selection(group):
    output = get_output(group)
    if "Inverted" not in output.inputs.keys():
        group.interface.new_socket(
            "Inverted", in_out="OUTPUT", socket_type="NodeSocketBool"
        )

    loc = output.location
    bool_math = group.nodes.new("FunctionNodeBooleanMath")
    bool_math.location = [loc[0], -100]
    bool_math.operation = "NOT"

    group.links.new(
        output.inputs["Selection"].links[0].from_socket, bool_math.inputs[0]
    )
    group.links.new(bool_math.outputs[0], output.inputs["Inverted"])


def boolean_link_output(tree: bpy.types.NodeTree, node: bpy.types.Node) -> None:
    link = tree.links.new
    node_output = get_output(tree)
    tree.interface.new_socket(
        name="Selection", in_out="OUTPUT", socket_type=socket_types["BOOLEAN"]
    )
    tree.interface.new_socket(
        name="Inverted", in_out="OUTPUT", socket_type=socket_types["BOOLEAN"]
    )
    final_output = node.outputs[0]
    link(final_output, node_output.inputs["Selection"])
    node_invert: bpy.types.FunctionNodeBooleanMath = tree.nodes.new(
        "FunctionNodeBooleanMath"
    )  # type: ignore

    node_invert.operation = "NOT"
    node_invert.location = (np.array(node_output.location) - [0, 200]).tolist()
    link(final_output, node_invert.inputs[0])
    link(node_invert.outputs[0], node_output.inputs["Inverted"])


def custom_iswitch(
    name: str,
    iter_list,
    field: str = "chain_id",
    dtype: str = "BOOLEAN",
    default_values=None,
    prefix: str = "",
    start: int = 0,
    offset: int = 0,
    panels: Optional[List[str]] = None,
    panels_open: int = 1,
):
    """
    Creates a named `Index Switch` node.

    Wraps an index switch node, giving the group names or each name in the `iter_list`. The
    inputs can also be placed in subpanels and given specific default values. Uses the
    given field for the attribute name to use in the index switch, and optionally adds an
    offset value if the start value is non zero.

    If a list of default items is given, then it is recycled to fill the defaults for
    each created socket in for the node.

    Parameters
    ----------
    name : str
        The name of the node group.
    iter_list : list
        The list of items to iterate over.
    field : str, optional
        The name of the attribute field. Defaults to 'chain_id'.
    default_values : list, optional
        The list of default values to assign to each item. Defaults to None.
    panels : list, str
        List of panel names for the sockets to be assigned to. If None, then socket will
        not be assigned to the panel. The will appear in the panel in the order in which
        they are given.
    panels_open: int
        Number of panels to default to open. If `0` then all panels will be closed. Useful
        for larger panel node groups, to keep them closed and help with organisation. A
        list of panel names can also be given to optionally categorise the sockets into
        panels in the final group node. The length of the panels list must match the length
        of the iter_list. `None` values mean the socket is not placed in a panel.
    prefix : str, optional
        The prefix to add to the node names. Defaults to an empty string.
    start : int, optional
        The starting index for the node names. Defaults to 0.

    Returns
    -------
    group : bpy.types.NodeGroup
        The created node group.

    Raises
    ------
    NodeGroupCreationError
        If there was an error creating the node group.
    """
    iter_list = [str(i) for i in iter_list]
    tree = bpy.data.node_groups.get(name)
    if tree:
        return tree

    socket_type = socket_types[dtype]
    tree = new_tree(name, geometry=False, fallback=False)

    # try creating the node group, otherwise on fail cleanup the created group and
    # report the error
    try:
        link = tree.links.new
        node_input = get_input(tree)
        node_output = get_output(tree)
        node_attr = tree.nodes.new("GeometryNodeInputNamedAttribute")
        node_attr.data_type = "INT"
        node_attr.location = [0, 150]
        node_attr.inputs["Name"].default_value = str(field)

        node_iswitch: bpy.types.GeometryNodeIndexSwitch = tree.nodes.new(  # type: ignore
            "GeometryNodeIndexSwitch"
        )
        node_iswitch.data_type = dtype

        link(node_attr.outputs["Attribute"], node_iswitch.inputs["Index"])

        # if there is as offset to the lookup values (say we want to start looking up
        # from 100 or 1000 etc) then we add a math node with that offset value
        if start != 0:
            node_math = tree.nodes.new("ShaderNodeMath")
            node_math.operation = "ADD"
            node_math.location = [0, 150]
            node_attr.location = [0, 300]

            node_math.inputs[1].default_value = start
            link(node_attr.outputs["Attribute"], node_math.inputs[0])
            link(node_math.outputs["Value"], node_iswitch.inputs["Index"])

        # if there are custom values provided, create a dictionary lookup for those values
        # to assign to the sockets upon creation. If no default was given and the dtype
        # is colors, then generate a random pastel color for each value
        default_lookup = None
        if default_values is not None:
            default_lookup = dict(zip(iter_list, itertools.cycle(default_values)))
        elif dtype == "RGBA":
            default_lookup = dict(
                zip(iter_list, [color.random_rgb() for i in iter_list])
            )

        # for each item in the iter_list, we create a new socket on the interface for this
        # node group, and link it to the interface on the index switch. The index switch
        # currently starts with two items already, so once i > 1 we start to add
        # new items for the index switch as well
        panel_item_counter = 0
        panel_counter = 0
        for j in range(offset):
            node_iswitch.index_switch_items.new()
        for i, item in enumerate(iter_list):
            if i > 1:
                node_iswitch.index_switch_items.new()

            # The offset creates but skips itmes on the index switch node.
            # if we offset by 1 then we can index from 1 essentially, for attributes like
            # the atomic_number

            socket = tree.interface.new_socket(
                name=f"{prefix}{item}", in_out="INPUT", socket_type=socket_type
            )
            #  if a set of default values was given, then use it for setting
            # the defaults on the created sockets of the node group
            if default_lookup:
                socket.default_value = default_lookup[item]
            link(
                node_input.outputs[socket.identifier],
                node_iswitch.inputs[str(i + offset)],
            )

            # if a list of panel names has been passed in, then we use it to
            # assign all of the interface sockets to the panels
            if panels:
                pname = panels[i]
                if pname is not None:
                    # try and get an existing panel, if None is returned we have to create
                    # a panel with the given name
                    panel: bpy.types.NodeTreeInterfacePanel = (
                        tree.interface.items_tree.get(pname)
                    )
                    if not panel:
                        panel_item_counter = 0
                        panel = tree.interface.new_panel(name=pname)
                        panel_counter += 1
                        # we can set a certain number of panels to be open when created.
                        # a value of 0 means all created panels will be closed on creation.
                        # larger cutoffs will mean that n number of panels will be open by
                        # default
                        if panel_counter >= panels_open:
                            panel.default_closed = True

                    tree.interface.move_to_parent(
                        socket, panel, to_position=panel_item_counter + 1
                    )
                    panel_item_counter += 1

        if dtype == "BOOLEAN":
            tree.color_tag = "INPUT"
            boolean_link_output(tree, node_iswitch)
        elif dtype == "RGBA":
            tree.color_tag = "COLOR"
            socket_out = tree.interface.new_socket(
                name="Color", in_out="OUTPUT", socket_type=socket_type
            )
            link(
                node_iswitch.outputs["Output"],
                node_output.inputs[socket_out.identifier],
            )
        else:
            raise ValueError(f"Unsupported value typee for custom iswitch: {dtype}")

        return tree

    # if something broke when creating the node group, delete whatever was created
    except Exception as e:
        node_name = tree.name
        bpy.data.node_groups.remove(tree)
        raise NodeGroupCreationError(
            f"Unable to make node group: {node_name}.\nError: {e}"
        )


def resid_multiple_selection(node_name, input_resid_string):
    """
    Returns a node group that takes an integer input and creates a boolean
    tick box for each item in the input list. Outputs are the selected
    residues and the inverse selection. Used for constructing chain
    selections in specific proteins.
    """

    # do a cleanning of input string to allow fuzzy input from users
    for c in ";/+ .":
        if c in input_resid_string:
            input_resid_string = input_resid_string.replace(c, ",")

    for c in "_=:":
        if c in input_resid_string:
            input_resid_string = input_resid_string.replace(c, "-")

    # parse input_resid_string into sub selecting string list
    sub_list = [item for item in input_resid_string.split(",") if item]

    # distance vertical to space all of the created nodes
    node_sep_dis = -100

    # get the active object, might need to change to taking an object as an input
    # and making it active isntead, to be more readily applied to multiple objects

    # create the custom node group data block, where everything will go
    # also create the required group node input and position it
    residue_id_group = bpy.data.node_groups.new(node_name, "GeometryNodeTree")
    node_input = residue_id_group.nodes.new("NodeGroupInput")
    node_input.location = [0, node_sep_dis * len(sub_list) / 2]

    group_link = residue_id_group.links.new
    new_node = residue_id_group.nodes.new

    prev = None
    for residue_id_index, residue_id in enumerate(sub_list):
        # add an new node of Select Res ID or MN_sek_res_id_range
        current_node = new_node("GeometryNodeGroup")

        # add an bool_math block
        bool_math = new_node("FunctionNodeBooleanMath")
        bool_math.location = [400, (residue_id_index + 1) * node_sep_dis]
        bool_math.operation = "OR"

        if "-" in residue_id:
            # selecting a range of residues by using the Res ID Range node and connecting
            # to the min and max of those nodes
            current_node.node_tree = append("Select Res ID Range")
            [resid_start, resid_end] = residue_id.split("-")[:2]
            socket_1 = residue_id_group.interface.new_socket(
                "res_id: Min", in_out="INPUT", socket_type="NodeSocketInt"
            )
            socket_1.default_value = int(resid_start)
            socket_2 = residue_id_group.interface.new_socket(
                "res_id: Max", in_out="INPUT", socket_type="NodeSocketInt"
            )
            socket_2.default_value = int(resid_end)

            # a residue range
            group_link(
                node_input.outputs[socket_1.identifier], current_node.inputs["Min"]
            )
            group_link(
                node_input.outputs[socket_2.identifier], current_node.inputs["Max"]
            )
        else:
            # Selecting singular res ID numbers by creating the socket and adding a node
            # ensuring that we are connecting to the right node
            current_node.node_tree = append("Select Res ID")
            socket = residue_id_group.interface.new_socket(
                "res_id", in_out="INPUT", socket_type="NodeSocketInt"
            )
            socket.default_value = int(residue_id)
            group_link(
                node_input.outputs[socket.identifier], current_node.inputs["Res ID"]
            )

        # set the coordinates
        current_node.location = [200, (residue_id_index + 1) * node_sep_dis]
        if not prev:
            # link the first residue selection to the first input of its OR block
            group_link(current_node.outputs["Selection"], bool_math.inputs[0])
        else:
            # if it is not the first residue selection, link the output to the previous or block
            group_link(current_node.outputs["Selection"], prev.inputs[1])

            # link the ouput of previous OR block to the current OR block
            group_link(prev.outputs[0], bool_math.inputs[0])
        prev = bool_math

    # add a output block
    residue_id_group_out = new_node("NodeGroupOutput")
    residue_id_group_out.location = [800, (residue_id_index + 1) / 2 * node_sep_dis]
    residue_id_group.interface.new_socket(
        "Selection", in_out="OUTPUT", socket_type="NodeSocketBool"
    )
    residue_id_group.interface.new_socket(
        "Inverted", in_out="OUTPUT", socket_type="NodeSocketBool"
    )
    group_link(prev.outputs[0], residue_id_group_out.inputs["Selection"])
    invert_bool_math = new_node("FunctionNodeBooleanMath")
    invert_bool_math.location = [600, (residue_id_index + 1) / 3 * 2 * node_sep_dis]
    invert_bool_math.operation = "NOT"
    group_link(prev.outputs[0], invert_bool_math.inputs[0])
    group_link(invert_bool_math.outputs[0], residue_id_group_out.inputs["Inverted"])
    return residue_id_group



import json
from pathlib import Path

import bpy
import numpy as np
from databpy import AttributeTypes, BlenderObject, store_named_attribute

from .base import Ensemble
from ... import blender as bl
from ... import color
from ..molecule.base import _create_object
from .reader import CellPackReader
from biotite.structure import AtomArray


class CellPack(Ensemble):
    def __init__(self, file_path):
        super().__init__(file_path)
        self.file_type = self._file_type()
        self.file = CellPackReader(file_path)
        self.file.get_molecules()
        self.transformations = self.file.assemblies(as_array=True)
        self.color_entity = {}
        self._color_palette_path = Path(file_path).parent / "color_palette.json"
        # self._setup_colors()

    def _setup_colors(self):
        if self._color_palette_path.exists():
            self.color_palette = json.load(open(self._color_palette_path))

        for entity in np.unique(self.array.entity_id):
            ename = self.data.entities[entity]
            if ename in self.color_palette:
                rgb = [self.color_palette[ename][c] / 255.0 for c in "xyz"]
                self.color_entity[entity] = np.array([*rgb, 1.0])
            else:
                self.color_entity[entity] = color.random_rgb(int(entity))

            self.entity_chains[entity] = (
                np.unique(self.array.asym_id[self.array.entity_id == entity]) @ property
            )

    @property
    def molecules(self):
        return self.file.molecules

    def create_object(
        self,
        name="CellPack",
        node_setup: bool = True,
        world_scale: float = 0.01,
        fraction: float = 1.0,
        simplify=False,
    ):
        self.object = self._create_data_object(name=f"{name}")
        self._create_object_instances(name=self.object.name, node_setup=node_setup)
        self._setup_node_tree(fraction=fraction)
        return self.object

    def _file_type(self):
        return Path(self.file_path).suffix.strip(".")

    def _assign_colors(self, obj: bpy.types.Object, array: AtomArray):
        # random color per chain
        # could also do by entity, + chain-lighten + atom-lighten

        entity = array.entity_id[0]
        color_entity = self.color_entity[entity]
        nc = len(self.entity_chains[entity])
        ci = np.where(self.entity_chains[entity] == chain_name)[0][0] * 2
        color_chain = color.Lab.lighten_color(color_entity, (float(ci) / nc))
        colors = np.tile(color_chain, (len(array), 1))

        store_named_attribute(
            obj=obj,
            name="Color",
            data=colors,
            atype=AttributeTypes.FLOAT_COLOR,
        )

    def _create_object_instances(
        self, name: str = "CellPack", node_setup: bool = True
    ) -> bpy.types.Collection:
        collection = bl.coll.cellpack(name)

        for i, mol_id in enumerate(self.file.mol_ids):
            array = self.molecules[mol_id]
            chain_name = array.asym_id[0]

            obj, coll_none = _create_object(
                array=array,
                name=mol_id,
                collection=collection,
            )

            if len(self.color_entity) > 0:
                self._assign_colors(obj, array)

            if node_setup:
                bl.nodes.create_starting_node_tree(
                    obj,
                    name=f"MN_pack_instance_{name}",
                    color=None,
                    material="MN Ambient Occlusion",
                )

        self.data_collection = collection
        self.instance_collection = collection

        return collection

    def _create_data_object(self, name="DataObject"):
        bob = BlenderObject(
            bl.mesh.create_data_object(
                self.transformations, name=name, collection=bl.coll.mn()
            )
        )
        bob.object["chain_ids"] = self.file.mol_ids

        # if we are dealing with petworld data, overwrite the chain_id for the data object
        if self.file._is_petworld:
            bob.store_named_attribute(bob.named_attribute("pdb_model_num"), "chain_id")

        return bob.object

    def _setup_node_tree(self, name="CellPack", fraction=1.0, as_points=False):
        mod = bl.nodes.get_mod(self.object)

        group = bl.nodes.new_tree(name=f"MN_ensemble_{name}", fallback=False)
        mod.node_group = group

        node_pack = bl.nodes.add_custom(group, "Ensemble Instance", location=[-100, 0])
        node_pack.inputs["Instances"].default_value = self.data_collection
        node_pack.inputs["Fraction"].default_value = fraction
        node_pack.inputs["As Points"].default_value = as_points

        link = group.links.new
        link(bl.nodes.get_input(group).outputs[0], node_pack.inputs[0])
        link(node_pack.outputs[0], bl.nodes.get_output(group).inputs[0])



"""
Importing molecular dynamics trajectories and associated files.
"""

__name__ = "MolecularNodes.trajectory"
__author__ = "Brady Johnston"

import bpy
import MDAnalysis as mda

from pathlib import Path
from ... import blender as bl
from ...style import STYLE_ITEMS
from ...session import MNSession
from .base import Trajectory
from . import dna
from bpy.props import StringProperty, EnumProperty, BoolProperty


def load(
    top: str | Path,
    traj: str | Path,
    name: str = "NewTrajectory",
    style: str | None = "spheres",
):
    top = bl.path_resolve(top)
    traj = bl.path_resolve(traj)

    universe = mda.Universe(top, traj)
    trajectory = Trajectory(universe=universe)
    trajectory.create_object(name=name, style=style)

    return trajectory


class MN_OT_Reload_Trajectory(bpy.types.Operator):
    bl_idname = "mn.reload_trajectory"
    bl_label = "Reload Trajectory"
    bl_description = (
        "Reload the `mda.UNiverse` of the current Object to renable updating"
    )
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        traj = context.scene.MNSession.match(obj)
        return not traj

    def execute(self, context):
        obj = context.active_object
        session: MNSession = context.scene.MNSession
        topo = obj.mn.filepath_topology
        traj = obj.mn.filepath_trajectory

        if "oxdna" in obj.mn.entity_type:
            uni = mda.Universe(
                topo, traj, topology_format=dna.OXDNAParser, format=dna.OXDNAReader
            )
            traj = dna.OXDNA(uni)
        else:
            uni = mda.Universe(topo, traj)
            traj = Trajectory(uni)

        traj.object = obj
        traj.set_frame(context.scene.frame_current)
        return {"FINISHED"}


class MN_OT_Import_Trajectory(bpy.types.Operator):
    bl_idname = "mn.import_trajectory"
    bl_label = "Import Protein MD"
    bl_description = "Load molecular dynamics trajectory"
    bl_options = {"REGISTER", "UNDO"}

    topology: StringProperty(  # type: ignore
        name="Topology",
        description="File path for the toplogy file for the trajectory",
        subtype="FILE_PATH",
        maxlen=0,
    )
    trajectory: StringProperty(  # type: ignore
        name="Trajectory",
        description="File path for the trajectory file for the trajectory",
        subtype="FILE_PATH",
        maxlen=0,
    )
    name: StringProperty(  # type: ignore
        name="Name",
        description="Name of the molecule on import",
        default="NewTrajectory",
        maxlen=0,
    )
    style: EnumProperty(  # type: ignore
        name="Style",
        description="Default style for importing",
        items=STYLE_ITEMS,
        default="spheres",
    )
    setup_nodes: BoolProperty(  # type: ignore
        name="Setup Nodes",
        description="Add nodes to the scene to load the trajectory",
        default=True,
    )

    def execute(self, context):
        trajectory = load(
            top=self.topology,
            traj=self.trajectory,
            name=self.name,
            style=self.style if self.setup_nodes else None,
        )

        context.view_layer.objects.active = trajectory.object
        context.scene.frame_start = 0
        context.scene.frame_end = trajectory.universe.trajectory.n_frames

        self.report(
            {"INFO"},
            message=f"Imported '{self.topology}' as {trajectory.name} "
            f"with {str(trajectory.universe.trajectory.n_frames)} "
            f"frames from '{self.trajectory}'.",
        )

        return {"FINISHED"}


# UI


def panel(layout, scene):
    layout.label(text="Load MD Trajectories", icon="FILE_TICK")
    layout.separator()
    col = layout.column(align=True)
    row_import = col.row()
    row_import.prop(scene.mn, "import_md_name")
    op = row_import.operator("mn.import_trajectory", text="Load")
    op.topology = scene.mn.import_md_topology
    op.trajectory = scene.mn.import_md_trajectory
    op.name = scene.mn.import_md_name
    op.style = scene.mn.import_style
    op.setup_nodes = scene.mn.import_node_setup
    col.separator()
    col.prop(scene.mn, "import_md_topology")
    col.prop(scene.mn, "import_md_trajectory")

    layout.separator()
    layout.label(text="Options", icon="MODIFIER")
    row = layout.row()
    row.prop(scene.mn, "import_node_setup", text="")
    col = row.column()
    col.prop(scene.mn, "import_style")
    col.enabled = scene.mn.import_node_setup


CLASSES = [MN_OT_Import_Trajectory, MN_OT_Reload_Trajectory]
